from pathlib import Path
from typing import TYPE_CHECKING, Optional

import click
import pytest
from _pytest._code.code import Traceback as PytestTraceback
from rich import print as rich_print

from ape.exceptions import ConfigError, ProviderNotConnectedError
from ape.logging import LogLevel, logger
from ape.pytest.utils import Scope
from ape.utils.basemodel import ManagerAccessMixin

if TYPE_CHECKING:
    from ape.api.networks import ProviderContextManager
    from ape.pytest.config import ConfigWrapper
    from ape.pytest.coverage import CoverageTracker
    from ape.pytest.fixtures import FixtureManager, IsolationManager, ReceiptCapture
    from ape.pytest.gas import GasTracker
    from ape.types.coverage import CoverageReport


class PytestApeRunner(ManagerAccessMixin):
    def __init__(
        self,
        config_wrapper: "ConfigWrapper",
        isolation_manager: "IsolationManager",
        receipt_capture: "ReceiptCapture",
        gas_tracker: "GasTracker",
        coverage_tracker: "CoverageTracker",
        fixture_manager: Optional["FixtureManager"] = None,
    ):
        self.config_wrapper = config_wrapper
        self.isolation_manager = isolation_manager
        self.receipt_capture = receipt_capture
        self._provider_is_connected = False

        # Ensure the gas report starts off None for this runner.
        gas_tracker.session_gas_report = None
        self.gas_tracker = gas_tracker
        self.coverage_tracker = coverage_tracker

        if fixture_manager is None:
            from ape.pytest.fixtures import FixtureManager

            self.fixture_manager = FixtureManager(config_wrapper, isolation_manager)

        else:
            self.fixture_manager = fixture_manager

        self._initialized_fixtures: list[str] = []
        self._finalized_fixtures: list[str] = []

    @property
    def _provider_context(self) -> "ProviderContextManager":
        return self.network_manager.parse_network_choice(self.config_wrapper.network)

    @property
    def _coverage_report(self) -> Optional["CoverageReport"]:
        return self.coverage_tracker.data.report if self.coverage_tracker.data else None

    def pytest_exception_interact(self, report, call):
        """
        A ``-I`` option triggers when an exception is raised which can be interactively handled.
        Outputs the full ``repr`` of the failed test and opens an interactive shell using the
        same console as the ``ape console`` command.
        """

        # Find the last traceback frame within the active project
        tb_frames: PytestTraceback = call.excinfo.traceback
        base = self.local_project.path.as_posix()

        if self.config_wrapper.show_internal:
            relevant_tb = list(tb_frames)
        else:
            relevant_tb = [
                f
                for f in tb_frames
                if Path(f.path).as_posix().startswith(base) or Path(f.path).name.startswith("test_")
            ]

        if relevant_tb:
            call.excinfo.traceback = PytestTraceback(relevant_tb)

            # Only show locals if not digging into the framework's traceback.
            # Else, it gets way too noisy.
            show_locals = not self.config_wrapper.show_internal

            try:
                here = Path.cwd()

            except FileNotFoundError:
                pass  # In a temp-folder, most likely.

            else:
                report.longrepr = call.excinfo.getrepr(
                    funcargs=True,
                    abspath=here,
                    showlocals=show_locals,
                    style="short",
                    tbfilter=False,
                    truncate_locals=True,
                    chain=False,
                )

        if self.config_wrapper.interactive and report.failed:
            from ape_console._cli import console

            traceback = call.excinfo.traceback[-1]

            # Suspend capsys to ignore our own output.
            capman = self.config_wrapper.get_pytest_plugin("capturemanager")
            if capman:
                capman.suspend_global_capture(in_=True)

            # Show the exception info before launching the interactive session.
            click.echo()
            rich_print(str(report.longrepr))
            click.echo()

            # get global namespace
            globals_dict = traceback.frame.f_globals

            # filter python internals and pytest internals
            globals_dict = {
                k: v
                for k, v in globals_dict.items()
                if not k.startswith("__") and not k.startswith("@")
            }

            # filter fixtures
            globals_dict = {
                k: v for k, v in globals_dict.items() if not hasattr(v, "_pytestfixturefunction")
            }

            # get local namespace
            locals_dict = traceback.locals
            locals_dict = {k: v for k, v in locals_dict.items() if not k.startswith("@")}

            click.echo("Starting interactive mode. Type `exit` to halt current test.")

            namespace = {"_callinfo": call, **globals_dict, **locals_dict}

            console(extra_locals=namespace, project=self.local_project, embed=True)

            if capman:
                capman.resume_global_capture()

        if type(call.excinfo.value) in (SystemExit, KeyboardInterrupt):
            # This will show the rest of Ape Test output as if the
            # tests had stopped here.
            pytest.exit("`ape test` exited.")

    def pytest_runtest_setup(self, item):
        """
        By default insert isolation fixtures into each test cases list of fixtures
        prior to actually executing the test case.

        https://docs.pytest.org/en/6.2.x/reference.html#pytest.hookspec.pytest_runtest_setup
        """
        if (
            not self.config_wrapper.isolation
            # doctests don't have fixturenames
            or (hasattr(pytest, "DoctestItem") and isinstance(item, pytest.DoctestItem))
            or "_function_isolation" in item.fixturenames  # prevent double injection
        ):
            # isolation is disabled via cmdline option or running doc-tests.
            return

        if self.config_wrapper.isolation:
            self._setup_isolation(item)

    def _setup_isolation(self, item):
        fixtures = self.fixture_manager.get_fixtures(item)
        for scope in (Scope.SESSION, Scope.PACKAGE, Scope.MODULE, Scope.CLASS):
            if not (
                custom_fixtures := [f for f in fixtures[scope] if self.fixture_manager.is_custom(f)]
            ):
                # Intermediate scope isolations aren't filled in, or only using
                # built-in Ape fixtures.
                continue

            snapshot = self.isolation_manager.get_snapshot(scope)

            # Gather new fixtures. Also, be mindful of parametrized fixtures
            # which strangely have the same name.
            new_fixtures = []
            for custom_fixture in custom_fixtures:
                # Parametrized fixtures must always be considered new
                # because of severe complications of using them.
                is_custom = custom_fixture in fixtures.parametrized
                is_iterating = is_custom and fixtures.is_iterating(custom_fixture)
                is_new = custom_fixture not in snapshot.fixtures

                # NOTE: Consider ``None`` to be stateful here to be safe.
                stateful = self.fixture_manager.is_stateful(custom_fixture) is not False

                if (is_new or is_iterating) and stateful:
                    new_fixtures.append(custom_fixture)
                    continue

            # Rebase if there are new fixtures found of non-function scope.
            # And there are stateful fixtures of lower scopes that need resetting.
            if self.fixture_manager.needs_rebase(new_fixtures, snapshot):
                self.fixture_manager.rebase(scope, fixtures)

            # Append these fixtures so we know when new ones arrive
            # and need to trigger the invalidation logic above.
            snapshot.append_fixtures(new_fixtures)

        fixtures.apply_fixturenames()

    def pytest_sessionstart(self):
        """
        Called after the `Session` object has been created and before performing
        collection and entering the run test loop.

        Removes `PytestAssertRewriteWarning` warnings from the terminalreporter.
        This prevents warnings that "the `ape` library was already imported and
        so related assertions cannot be rewritten". The warning is not relevant
        for end users who are performing tests with ape.
        """
        reporter = self.config_wrapper.get_pytest_plugin("terminalreporter")
        if not reporter:
            return

        warnings = reporter.stats.pop("warnings", [])
        warnings = [i for i in warnings if "PytestAssertRewriteWarning" not in i.message]
        if warnings and not self.config_wrapper.disable_warnings:
            reporter.stats["warnings"] = warnings

    @pytest.hookimpl(tryfirst=True, hookwrapper=True)
    def pytest_runtest_call(self, item):
        if network_marker := item.get_closest_marker("use_network"):
            if len(getattr(network_marker, "args", []) or []) != 1:
                raise ValueError("`use_network` marker requires single network choice argument.")

            with self.network_manager.parse_network_choice(network_marker.args[0]):
                yield

        else:
            yield

    def pytest_fixture_setup(self, fixturedef, request):
        fixture_name = fixturedef.argname
        if fixture_name in self._initialized_fixtures:
            return

        self._initialized_fixtures.append(fixture_name)
        if self._track_fixture_blocks(fixture_name):
            try:
                block_number = self.chain_manager.blocks.height
            except Exception:
                pass
            else:
                self.fixture_manager.add_fixture_info(fixture_name, setup_block=block_number)

    def pytest_fixture_post_finalizer(self, fixturedef, request):
        fixture_name = fixturedef.argname
        if fixture_name in self._finalized_fixtures:
            return

        self._finalized_fixtures.append(fixture_name)
        if self._track_fixture_blocks(fixture_name):
            try:
                block_number = self.chain_manager.blocks.height
            except ProviderNotConnectedError:
                pass
            else:
                self.fixture_manager.add_fixture_info(fixture_name, teardown_block=block_number)

    def _track_fixture_blocks(self, fixture_name: str) -> bool:
        if not self.fixture_manager.is_custom(fixture_name):
            return False

        scope = self.fixture_manager.get_fixture_scope(fixture_name)
        if scope in (None, Scope.FUNCTION):
            return False

        return True

    @pytest.hookimpl(trylast=True, hookwrapper=True)
    def pytest_collection_finish(self, session):
        """
        Called after collection has been performed and modified.
        """
        outcome = yield

        # Only start provider if collected tests.
        if not outcome.get_result() and session.items:
            self._connect()

    def _connect(self):
        if self._provider_context._provider.network.is_mainnet:
            # Ensure is not only running on tests on mainnet because
            # was configured as the default.
            is_from_command_line = (
                "--network" in self.config_wrapper.pytest_config.invocation_params.args
            )
            if not is_from_command_line:
                raise ConfigError(
                    "Default network is mainnet; unable to run tests on mainnet. "
                    "Please specify the network using the `--network` flag or "
                    "configure a different default network."
                )

        self._provider_context.push_provider()
        self._provider_is_connected = True

    def pytest_terminal_summary(self, terminalreporter):
        """
        Add a section to terminal summary reporting.
        When ``--gas`` is active, outputs the gas profile report.
        """
        if self.config_wrapper.track_gas:
            self._show_gas_report(terminalreporter)
        if self.config_wrapper.track_coverage:
            self._show_coverage_report(terminalreporter)

    def _show_gas_report(self, terminalreporter):
        terminalreporter.section("Gas Profile")
        if not self.network_manager.connected:
            # Happens if never needed to connect (no tests)
            return

        self._log_tracing_support(
            terminalreporter, "The gas profile is limited to receipt-level data."
        )
        if not self.gas_tracker.show_session_gas():
            terminalreporter.write_line(
                f"{LogLevel.WARNING.name}: No gas usage data found.", yellow=True
            )

    def _show_coverage_report(self, terminalreporter):
        if self.config_wrapper.ape_test_config.coverage.reports.terminal:
            terminalreporter.section("Coverage Profile")

        if not self.network_manager.connected:
            # Happens if never needed to connect (no tests)
            return

        self._log_tracing_support(
            terminalreporter, "Coverage is limited to receipt-level function coverage."
        )
        if not self.coverage_tracker.show_session_coverage():
            terminalreporter.write_line(
                f"{LogLevel.WARNING.name}: No coverage data found. "
                f"Try re-compiling your contracts using the latest compiler plugins",
                yellow=True,
            )

    def _log_tracing_support(self, terminalreporter, extra_warning: str):
        if self.provider.supports_tracing:
            return

        message = (
            f"{LogLevel.ERROR.name}: Provider '{self.provider.name}' does not support "
            f"transaction tracing. {extra_warning}"
        )
        terminalreporter.write_line(message, red=True)

    def pytest_unconfigure(self):
        if self._provider_is_connected and self.config_wrapper.disconnect_providers_after:
            try:
                self._provider_context.disconnect_all()
            except Exception as err:
                logger.error(f"Failed to disconnect {self}: {err}")
            else:
                self._provider_is_connected = False

        # NOTE: Clearing the state is helpful for pytester-based tests,
        #  which may run pytest many times in-process.
        self.receipt_capture.clear()
        self.chain_manager.contracts.clear_local_caches()
        self.gas_tracker.session_gas_report = None
        self.coverage_tracker.reset()



import threading
import time
from contextlib import contextmanager
from pathlib import Path
from shutil import copytree
from typing import TYPE_CHECKING, Optional, cast

import pytest
from eth_pydantic_types import HexBytes
from eth_utils import to_hex
from ethpm_types import ContractType, ErrorABI, MethodABI
from ethpm_types.abi import ABIType

import ape
from ape.contracts import ContractContainer, ContractInstance
from ape.contracts.base import ContractCallHandler
from ape.exceptions import ChainError, ContractLogicError, ProviderError
from ape.logging import LogLevel
from ape.logging import logger as _logger
from ape.types.address import AddressType
from ape.utils.misc import LOCAL_NETWORK_NAME
from ape_ethereum.proxies import minimal_proxy as _minimal_proxy_container

if TYPE_CHECKING:
    from ape.types.events import ContractLog


ALIAS_2 = "__FUNCTIONAL_TESTS_ALIAS_2__"
TEST_ADDRESS = cast(AddressType, "0xd8dA6BF26964aF9D7eEd9e03E53415D37aA96045")
BASE_PROJECTS_DIRECTORY = (Path(__file__).parent / "data" / "projects").absolute()
PROJECT_WITH_LONG_CONTRACTS_FOLDER = BASE_PROJECTS_DIRECTORY / "LongContractsFolder"
APE_PROJECT_FOLDER = BASE_PROJECTS_DIRECTORY / "ApeProject"
BASE_SOURCES_DIRECTORY = (Path(__file__).parent / "data/sources").absolute()

CALL_WITH_STRUCT_INPUT = MethodABI.model_validate(
    {
        "type": "function",
        "name": "getTradeableOrderWithSignature",
        "stateMutability": "view",
        "inputs": [
            {"name": "owner", "type": "address", "internalType": "address"},
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {
                        "name": "handler",
                        "type": "address",
                        "internalType": "contract IConditionalOrder",
                    },
                    {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                ],
                "internalType": "struct IConditionalOrder.ConditionalOrderParams",
            },
            {"name": "offchainInput", "type": "bytes", "internalType": "bytes"},
            {"name": "proof", "type": "bytes32[]", "internalType": "bytes32[]"},
        ],
        "outputs": [
            {
                "name": "order",
                "type": "tuple",
                "components": [
                    {"name": "sellToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "buyToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "receiver", "type": "address", "internalType": "address"},
                    {"name": "sellAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "buyAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "validTo", "type": "uint32", "internalType": "uint32"},
                    {"name": "appData", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "feeAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "kind", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "partiallyFillable", "type": "bool", "internalType": "bool"},
                    {"name": "sellTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "buyTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                ],
                "internalType": "struct GPv2Order.Data",
            },
            {"name": "signature", "type": "bytes", "internalType": "bytes"},
        ],
    }
)
METHOD_WITH_STRUCT_INPUT = MethodABI.model_validate(
    {
        "type": "function",
        "name": "getTradeableOrderWithSignature",
        "stateMutability": "view",
        "inputs": [
            {"name": "owner", "type": "address", "internalType": "address"},
            {
                "name": "params",
                "type": "tuple",
                "components": [
                    {
                        "name": "handler",
                        "type": "address",
                        "internalType": "contract IConditionalOrder",
                    },
                    {"name": "salt", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "staticInput", "type": "bytes", "internalType": "bytes"},
                ],
                "internalType": "struct IConditionalOrder.ConditionalOrderParams",
            },
            {"name": "offchainInput", "type": "bytes", "internalType": "bytes"},
            {"name": "proof", "type": "bytes32[]", "internalType": "bytes32[]"},
        ],
        "outputs": [
            {
                "name": "order",
                "type": "tuple",
                "components": [
                    {"name": "sellToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "buyToken", "type": "address", "internalType": "contract IERC20"},
                    {"name": "receiver", "type": "address", "internalType": "address"},
                    {"name": "sellAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "buyAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "validTo", "type": "uint32", "internalType": "uint32"},
                    {"name": "appData", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "feeAmount", "type": "uint256", "internalType": "uint256"},
                    {"name": "kind", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "partiallyFillable", "type": "bool", "internalType": "bool"},
                    {"name": "sellTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                    {"name": "buyTokenBalance", "type": "bytes32", "internalType": "bytes32"},
                ],
                "internalType": "struct GPv2Order.Data",
            },
            {"name": "signature", "type": "bytes", "internalType": "bytes"},
        ],
    }
)


class _ContractLogicError(ContractLogicError):
    pass


@pytest.hookimpl(trylast=True, hookwrapper=True)
def pytest_collection_finish(session):
    with ape.networks.parse_network_choice("::test"):
        # Sets the active provider
        yield


@pytest.fixture
def mock_web3(mocker):
    return mocker.MagicMock()


@pytest.fixture
def mock_transaction(mocker):
    tx = mocker.MagicMock()
    tx.required_confirmations = 0
    return tx


@pytest.fixture(scope="session")
def address():
    return TEST_ADDRESS


@pytest.fixture
def second_keyfile_account(sender, keyparams, temp_keyfile_account_ctx):
    with temp_keyfile_account_ctx(ALIAS_2, keyparams, sender) as account:
        # Ensure starts off locked.
        account.lock()
        yield account


@pytest.fixture
def solidity_contract_instance(
    owner, solidity_contract_container, networks_connected_to_tester
) -> ContractInstance:
    return owner.deploy(solidity_contract_container, 0)


@pytest.fixture(scope="session")
def solidity_fallback_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("SolFallbackAndReceive")


@pytest.fixture(scope="session")
def vyper_fallback_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("VyDefault")


@pytest.fixture(scope="session")
def solidity_fallback_container(solidity_fallback_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=solidity_fallback_contract_type)


@pytest.fixture(scope="session")
def vyper_fallback_container(vyper_fallback_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=vyper_fallback_contract_type)


@pytest.fixture
def vyper_contract_instance(
    owner, vyper_contract_container, networks_connected_to_tester
) -> ContractInstance:
    return owner.deploy(vyper_contract_container, 0, required_confirmations=0)


@pytest.fixture
def solidity_fallback_contract(owner, solidity_fallback_container):
    return owner.deploy(solidity_fallback_container)


@pytest.fixture
def vyper_fallback_contract(owner, vyper_fallback_container):
    return owner.deploy(vyper_fallback_container)


@pytest.fixture(scope="session")
def reverts_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("RevertsContract")


@pytest.fixture(scope="session")
def sub_reverts_contract_type(get_contract_type) -> ContractType:
    return get_contract_type("SubReverts")


@pytest.fixture(scope="session")
def reverts_contract_container(reverts_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=reverts_contract_type)


@pytest.fixture(scope="session")
def sub_reverts_contract_container(sub_reverts_contract_type) -> ContractContainer:
    return ContractContainer(contract_type=sub_reverts_contract_type)


@pytest.fixture
def reverts_contract_instance(
    owner, reverts_contract_container, sub_reverts_contract_instance, eth_tester_provider
) -> ContractInstance:
    return owner.deploy(
        reverts_contract_container, sub_reverts_contract_instance, required_confirmations=0
    )


@pytest.fixture(params=("solidity", "vyper"))
def contract_container(
    request, solidity_contract_container, vyper_contract_container, networks_connected_to_tester
):
    return solidity_contract_container if request.param == "solidity" else vyper_contract_container


@pytest.fixture(params=("solidity", "vyper"))
def contract_instance(
    eth_tester_provider, request, solidity_contract_instance, vyper_contract_instance
):
    return solidity_contract_instance if request.param == "solidity" else vyper_contract_instance


@pytest.fixture(params=("solidity", "vyper"))
def fallback_contract(
    eth_tester_provider, request, solidity_fallback_contract, vyper_fallback_contract
):
    return solidity_fallback_contract if request.param == "solidity" else vyper_fallback_contract


@pytest.fixture
def ds_note_test_contract(eth_tester_provider, vyper_contract_type, owner, get_contract_type):
    contract_type = get_contract_type("DsNoteTest")
    contract_container = ContractContainer(contract_type=contract_type)
    return contract_container.deploy(sender=owner)


@pytest.fixture(scope="session")
def project_with_contract():
    with ape.Project(APE_PROJECT_FOLDER).isolate_in_tempdir() as project:
        yield project


@pytest.fixture(scope="session")
def project_with_source_files_contract(project_with_contract):
    bases_source_dir = BASE_SOURCES_DIRECTORY
    project_source_dir = project_with_contract.path

    with ape.Project.create_temporary_project() as tmp_project:
        copytree(project_source_dir, str(tmp_project.path), dirs_exist_ok=True)
        copytree(bases_source_dir, tmp_project.path / "contracts", dirs_exist_ok=True)
        yield tmp_project


@pytest.fixture
def clean_contract_caches(chain):
    with chain.contracts.use_temporary_caches():
        yield


@pytest.fixture
def project_with_dependency_config(project):
    dependencies_config = {
        "contracts_folder": "functional/data/contracts/local",
        "dependencies": [
            {
                "local": str(PROJECT_WITH_LONG_CONTRACTS_FOLDER),
                "name": "testdependency",
                "config_override": {
                    "contracts_folder": "source/v0.1",
                },
                "version": "releases/v6",  # Testing having a slash in version.
            }
        ],
    }
    project.clean()
    with project.isolate_in_tempdir(**dependencies_config) as tmp_project:
        yield tmp_project


@pytest.fixture(scope="session")
def base_projects_directory():
    return BASE_PROJECTS_DIRECTORY


@pytest.fixture(scope="session")
def mainnet_contract(chain):
    def contract_getter(address):
        path = (
            Path(__file__).parent
            / "data"
            / "contracts"
            / "ethereum"
            / "mainnet"
            / f"{address}.json"
        )
        contract = ContractType.model_validate_json(path.read_text())
        chain.contracts[address] = contract
        return contract

    return contract_getter


@pytest.fixture(scope="session")
def ds_note():
    return {
        "address": "0x35D1b3F3D7966A1DFe207aa4514C12a259A0492B",
        "topics": [
            HexBytes("0x7608870300000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x5946492d41000000000000000000000000000000000000000000000000000000"),
            HexBytes("0x0000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a45"),
            HexBytes("0x0000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a45"),
        ],
        "data": "0x000000000000000000000000000000000000000000000000000000000000002000000000000000000000000000000000000000000000000000000000000000e0760887035946492d410000000000000000000000000000000000000000000000000000000000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000abb839063ef747c8432b2acc60bf8f70ec09a450000000000000000000000000000000000000000000000000000000000000000fffffffffffffffffffffffffffffffffffffffffffa050e82a57b7fc6b6020c00000000000000000000000000000000000000000000000000000000",
        "blockNumber": 14623434,
        "transactionHash": HexBytes(
            "0xa322a9fd0e627e22bfe1b0877cca1d1f2e697d076007231d0b7a366d1a0fdd51"
        ),
        "transactionIndex": 333,
        "blockHash": HexBytes("0x0fd77b0af3fa471aa040a02d4fcd1ec0a35122a4166d0bb7c31354e23823de49"),
        "logIndex": 376,
        "removed": False,
    }


@pytest.fixture
def chain_that_mined_5(chain):
    chain.mine(5)
    return chain


class PollDaemonThread(threading.Thread):
    def __init__(self, name, poller, handler, stop_condition, *args, **kwargs):
        kwargs_dict = dict(**kwargs)
        kwargs_dict["name"] = f"ape_poll_{name}"
        super().__init__(*args, **kwargs_dict)
        self._poller = poller
        self._handler = handler
        self._do_stop = stop_condition
        self._exception = None
        self._max_iterations = 100

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.stop()

    def run(self):
        try:
            self._run_until_stop(timeout_iterations=self._max_iterations)
        except Exception as err:
            self._exception = err

    def stop(self):
        time.sleep(1)
        self.join()

        # Attempt to wait for stop condition
        if not self._do_stop():
            self._run_until_stop(timeout_iterations=10)

    def join(self, timeout=None):
        super().join(timeout=timeout)
        if self._exception and not self._do_stop():
            # Only raise if error-ed before hitting stop condition
            raise self._exception

    def _run_until_stop(self, timeout_iterations: Optional[int] = None):
        iterations = 0
        while True:
            if self._do_stop():
                return

            try:
                self._handler(next(self._poller))
            except (ChainError, ProviderError):
                # Check if we can stop once more before exiting
                if self._do_stop():
                    return

                raise  # The timeout ChainError

            time.sleep(1)
            if timeout_iterations is None:
                continue

            elif iterations >= timeout_iterations:
                return

            iterations += 1


@pytest.fixture
def PollDaemon():
    return PollDaemonThread


@pytest.fixture
def assert_log_values(contract_instance):
    def _assert_log_values(
        log: "ContractLog",
        number: int,
        previous_number: Optional[int] = None,
        address: Optional[AddressType] = None,
    ):
        assert log.contract_address == address or contract_instance.address
        assert isinstance(log.b, bytes)
        expected_previous_number = number - 1 if previous_number is None else previous_number
        assert log.prevNum == expected_previous_number, "Event param 'prevNum' has unexpected value"
        assert log.newNum == number, "Event param 'newNum' has unexpected value"
        assert log.dynData == "Dynamic"
        assert log.dynIndexed == HexBytes(
            "0x9f3d45ac20ccf04b45028b8080bb191eab93e29f7898ed43acf480dd80bba94d"
        )

    return _assert_log_values


@pytest.fixture(scope="session")
def logger():
    _logger.set_level(LogLevel.ERROR)
    return _logger


@pytest.fixture
def use_debug(logger):
    initial_level = logger.level
    logger.set_level(LogLevel.DEBUG)
    yield
    logger.set_level(initial_level)


@pytest.fixture
def dummy_live_network(chain):
    original_network = chain.provider.network.name
    chain.provider.network.name = "sepolia"
    yield chain.provider.network
    chain.provider.network.name = original_network


@pytest.fixture
def dummy_live_network_with_explorer(dummy_live_network, mock_explorer):
    dummy_live_network.__dict__["explorer"] = mock_explorer
    yield dummy_live_network
    dummy_live_network.__dict__.pop("explorer", None)


@pytest.fixture(scope="session")
def proxy_contract_container(get_contract_type):
    return ContractContainer(get_contract_type("proxy"))


@pytest.fixture(scope="session")
def calldata():
    return HexBytes(
        "0x3fb5c1cb00000000000000000000000000000000000000000000000000000000000000de"
    )  # setNumber(222)


@pytest.fixture(scope="session")
def calldata_with_address():
    return HexBytes(
        "0x9e6b154b00000000000000000000000000000000000000000000000000000000000000de"
        "000000000000000000000000f7f78379391c5df2db5b66616d18ff92edb82022"
    )  # setNumber(222, 0xf7f78379391c5df2db5b66616d18ff92edb82022)


@pytest.fixture(scope="session")
def unique_calldata():
    return HexBytes(
        "0xacab48d8"
        "0000000000000000000000000000000000000000000000000000000000000000"
        "0000000000000000000000000000000000000000000000000000000000000001"
        "0000000000000000000000000000000000000000000000000000000000000002"
        "0000000000000000000000000000000000000000000000000000000000000003"
        "0000000000000000000000000000000000000000000000000000000000000004"
        "0000000000000000000000000000000000000000000000000000000000000005"
        "0000000000000000000000000000000000000000000000000000000000000006"
        "0000000000000000000000000000000000000000000000000000000000000007"
        "0000000000000000000000000000000000000000000000000000000000000008"
        "0000000000000000000000000000000000000000000000000000000000000009"
    )  # functionWithUniqueArguments(0, 1, 2, 3, 4, 5, 6, 7, 8, 9)


@pytest.fixture
def leaf_contract(eth_tester_provider, owner, get_contract_type):
    ct = get_contract_type("ContractC")
    return owner.deploy(ContractContainer(ct))


@pytest.fixture
def middle_contract(eth_tester_provider, owner, get_contract_type, leaf_contract):
    ct = get_contract_type("ContractB")
    return owner.deploy(ContractContainer(ct), leaf_contract)


@pytest.fixture
def contract_with_call_depth(
    owner, eth_tester_provider, get_contract_type, leaf_contract, middle_contract
):
    contract = ContractContainer(get_contract_type("ContractA"))
    return owner.deploy(contract, middle_contract, leaf_contract)


@pytest.fixture
def sub_reverts_contract_instance(owner, sub_reverts_contract_container, eth_tester_provider):
    return owner.deploy(sub_reverts_contract_container, required_confirmations=0)


@pytest.fixture(scope="session")
def error_contract_container(get_contract_type):
    ct = get_contract_type("HasError")
    return ContractContainer(ct)


@pytest.fixture
def error_contract(owner, error_contract_container, eth_tester_provider):
    _ = eth_tester_provider  # Ensure uses eth tester
    return owner.deploy(error_contract_container, 1)


@pytest.fixture
def vyper_factory(owner, get_contract_type):
    return owner.deploy(ContractContainer(get_contract_type("VyperFactory")))


@pytest.fixture
def vyper_printing(owner, get_contract_type):
    return owner.deploy(ContractContainer(get_contract_type("printing")))


@pytest.fixture
def vyper_blueprint(owner, vyper_contract_container):
    receipt = owner.declare(vyper_contract_container)
    return receipt.contract_address


@pytest.fixture
def minimal_proxy_container():
    return _minimal_proxy_container


@pytest.fixture
def minimal_proxy(owner, minimal_proxy_container):
    return owner.deploy(minimal_proxy_container)


@pytest.fixture
def mock_explorer(mocker):
    explorer = mocker.MagicMock()
    explorer.name = "mock"  # Needed for network data serialization.
    return explorer


@pytest.fixture
def call_abi_with_struct_input():
    return CALL_WITH_STRUCT_INPUT


@pytest.fixture
def fake_contract(mocker):
    # Only needed for initialization; never used.
    return mocker.MagicMock()


@pytest.fixture
def call_handler_with_struct_input(fake_contract, call_abi_with_struct_input):
    abi = call_abi_with_struct_input
    return ContractCallHandler(contract=fake_contract, abis=[abi])


@pytest.fixture(scope="session")
def struct_input_for_call(owner):
    return [owner, [owner, b"skip", b"skip"], b"skip", [b"skip"]]


@pytest.fixture(scope="session")
def output_from_struct_input_call(accounts):
    # Expected when using `struct_input_for_call`.
    addr = accounts[0].address.replace("0x", "")

    return HexBytes(
        f"0x26e0a196000000000000000000000000{addr}000000000000000000000000000000000"
        f"0000000000000000000000000000080000000000000000000000000000000000000000000"
        f"0000000000000000000120000000000000000000000000000000000000000000000000000"
        f"0000000000160000000000000000000000000{addr}736b69700000000000000000000000"
        f"0000000000000000000000000000000000000000000000000000000000000000000000000"
        f"0000000000000000000000060000000000000000000000000000000000000000000000000"
        f"0000000000000004736b69700000000000000000000000000000000000000000000000000"
        f"0000000000000000000000000000000000000000000000000000000000000000000000473"
        f"6b69700000000000000000000000000000000000000000000000000000000000000000000"
        f"00000000000000000000000000000000000000000000000000001736b6970000000000000"
        f"00000000000000000000000000000000000000000000"
    )


@pytest.fixture
def method_abi_with_struct_input():
    return METHOD_WITH_STRUCT_INPUT


@pytest.fixture
def mock_compiler(make_mock_compiler):
    return make_mock_compiler()


@pytest.fixture
def make_mock_compiler(mocker):
    def fn(name="mock"):
        mock = mocker.MagicMock()
        mock.name = "mock"
        mock.ext = f".__{name}__"
        mock.tracked_settings = []
        mock.ast = None
        mock.pcmap = None
        mock.abi = []

        def mock_compile(paths, project=None, settings=None):
            settings = settings or {}
            mock.tracked_settings.append(settings)
            result = []
            for path in paths:
                if path.suffix == mock.ext:
                    name = path.stem
                    code = to_hex(123)
                    data = {
                        "contractName": name,
                        "abi": mock.abi,
                        "deploymentBytecode": code,
                        "sourceId": f"{project.contracts_folder.name}/{path.name}",
                    }
                    if ast := mock.ast:
                        data["ast"] = ast
                    if pcmap := mock.pcmap:
                        data["pcmap"] = pcmap

                    # Check for mocked overrides
                    overrides = mock.overrides
                    if isinstance(overrides, dict):
                        data = {**data, **overrides}

                    contract_type = ContractType.model_validate(data)
                    result.append(contract_type)

            return result

        mock.compile.side_effect = mock_compile
        return mock

    return fn


@pytest.fixture
def mock_sepolia(create_mock_sepolia):
    """
    Temporarily tricks Ape into thinking the local network
    is Sepolia so we can test features that require a live
    network.
    """
    with create_mock_sepolia() as network:
        yield network


@pytest.fixture
def create_mock_sepolia(ethereum, eth_tester_provider, vyper_contract_instance):
    @contextmanager
    def fn():
        # Ensuring contract exists before hack.
        # This allows the network to be past genesis which is more realistic.
        _ = vyper_contract_instance
        eth_tester_provider.network.name = "sepolia"
        yield eth_tester_provider.network
        eth_tester_provider.network.name = LOCAL_NETWORK_NAME

    return fn


@pytest.fixture
def disable_fork_providers(ethereum):
    """
    When ape-hardhat or ape-foundry is installed,
    this tricks the test into thinking they are not
    (only uses sepolia-fork).
    """
    actual = ethereum.sepolia_fork.__dict__.pop("providers", {})
    ethereum.sepolia_fork.__dict__["providers"] = {}
    yield
    if actual:
        ethereum.sepolia_fork.__dict__["providers"] = actual


@pytest.fixture
def mock_fork_provider(mocker, ethereum, mock_sepolia):
    """
    A fake provider representing something like ape-foundry
    that can fork networks (only uses sepolia-fork).
    """
    initial_providers = ethereum.sepolia_fork.__dict__.pop("providers", {})
    initial_default = ethereum.sepolia_fork._default_provider
    mock_provider = mocker.MagicMock()
    mock_provider.name = "mock"
    mock_provider.network = ethereum.sepolia_fork

    # Have to do this because providers are partials.
    def fake_partial(*args, **kwargs):
        mock_provider.partial_call = (args, kwargs)
        return mock_provider

    ethereum.sepolia_fork._default_provider = "mock"
    ethereum.sepolia_fork.__dict__["providers"] = {"mock": fake_partial}
    yield mock_provider
    if initial_providers:
        ethereum.sepolia_fork.__dict__["providers"] = initial_providers
    if initial_default:
        ethereum.sepolia_fork._default_provider = initial_default


@pytest.fixture
def delete_account_after():
    @contextmanager
    def delete_account_context(alias: str):
        yield
        account_path = ape.config.DATA_FOLDER / "accounts" / f"{alias}.json"
        if account_path.is_file():
            account_path.unlink()

    return delete_account_context


@pytest.fixture
def setup_custom_error(chain):
    def fn(addr: AddressType):
        abi = [
            ErrorABI(
                type="error",
                name="AllowanceExpired",
                inputs=[
                    ABIType(
                        name="deadline", type="uint256", components=None, internal_type="uint256"
                    )
                ],
            ),
            MethodABI(
                type="function",
                name="execute",
                stateMutability="payable",
                inputs=[
                    ABIType(name="commands", type="bytes", components=None, internal_type="bytes"),
                    ABIType(
                        name="inputs", type="bytes[]", components=None, internal_type="bytes[]"
                    ),
                ],
                outputs=[],
            ),
        ]
        contract_type = ContractType(abi=abi)

        # Hack in contract-type.
        chain.contracts.contract_types[addr] = contract_type

    return fn



from typing import TYPE_CHECKING, Any, Optional

from pydantic_core.core_schema import (
    CoreSchema,
    ValidationInfo,
    int_schema,
    no_info_plain_validator_function,
    plain_serializer_function_ser_schema,
)

from ape.exceptions import ConversionError
from ape.utils.basemodel import ManagerAccessMixin

if TYPE_CHECKING:
    from typing_extensions import TypeAlias


class CurrencyValueComparable(int):
    """
    An integer you can compare with currency-value
    strings, such as ``"1 ether"``.
    """

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, int):
            return super().__eq__(other)

        elif isinstance(other, str):
            try:
                other_value = ManagerAccessMixin.conversion_manager.convert(other, int)
            except ConversionError:
                # Not a currency-value, it's ok.
                return False

            return super().__eq__(other_value)

        # Try from the other end, if hasn't already.
        return NotImplemented

    def __hash__(self) -> int:
        return hash(int(self))

    @classmethod
    def __get_pydantic_core_schema__(cls, value, handler=None) -> CoreSchema:
        return no_info_plain_validator_function(
            cls._validate,
            serialization=plain_serializer_function_ser_schema(
                cls._serialize,
                info_arg=False,
                return_schema=int_schema(),
            ),
        )

    @staticmethod
    def _validate(value: Any, info: Optional[ValidationInfo] = None) -> "CurrencyValueComparable":
        # NOTE: For some reason, for this to work, it has to happen
        #   in an "after" validator, or else it always only `int` type on the model.
        if value is None:
            # Will fail if not optional.
            # Type ignore because this is an hacky and unlikely situation.
            return None  # type: ignore

        elif isinstance(value, str) and " " in value:
            return ManagerAccessMixin.conversion_manager.convert(value, int)

        # For models annotating with this type, we validate all integers into it.
        return CurrencyValueComparable(value)

    @staticmethod
    def _serialize(value):
        return int(value)


CurrencyValueComparable.__name__ = int.__name__


CurrencyValue: "TypeAlias" = CurrencyValueComparable
"""
An alias to :class:`~ape.types.CurrencyValueComparable` for
situations when you know for sure the type is a currency-value
(and not just comparable to one).
"""



import warnings
from collections.abc import Iterator
from typing import TYPE_CHECKING, Any, Optional, Union, cast

from eip712.messages import EIP712Message
from eth_account import Account as EthAccount
from eth_account._utils.signing import sign_transaction_dict
from eth_account.messages import SignableMessage, encode_defunct
from eth_keys.datatypes import PrivateKey  # type: ignore
from eth_pydantic_types import HexBytes
from eth_utils import to_bytes, to_canonical_address, to_hex

from ape.api.accounts import TestAccountAPI, TestAccountContainerAPI
from ape.api.address import BaseAddress
from ape.exceptions import APINotImplementedError, ProviderNotConnectedError, SignatureError
from ape.types import AddressType
from ape.types.signatures import MessageSignature, TransactionSignature
from ape.utils._web3_compat import sign_hash
from ape.utils.misc import ZERO_ADDRESS, derive_public_key, log_instead_of_fail
from ape.utils.testing import generate_dev_accounts
from ape_ethereum import Authorization
from ape_ethereum.transactions import TransactionType

if TYPE_CHECKING:
    from ape.api.transactions import TransactionAPI


class TestAccountContainer(TestAccountContainerAPI):
    generated_accounts: list["TestAccount"] = []

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __len__(self) -> int:
        return self.number_of_accounts + len(self.generated_accounts)

    @property
    def mnemonic(self) -> str:
        # Overridden so we can overload the setter.
        return self.config_manager.test.mnemonic

    @mnemonic.setter
    def mnemonic(self, mnemonic: str) -> None:
        # Overridden so we can also clear out generated accounts cache.
        self.config_manager.test.mnemonic = mnemonic
        self.generated_accounts = []

    @TestAccountContainerAPI.mnemonic.setter
    def mnemonic(self, mnemonic: str) -> None:
        self.config_manager.test.mnemonic = mnemonic
        self.generated_accounts = []

    @property
    def config(self):
        return self.config_manager.get_config("test")

    @property
    def aliases(self) -> Iterator[str]:
        for index in range(self.number_of_accounts):
            yield f"TEST::{index}"

    @property
    def accounts(self) -> Iterator["TestAccount"]:
        for index in range(self.number_of_accounts):
            yield cast(TestAccount, self.get_test_account(index))

    def get_test_account(self, index: int) -> TestAccountAPI:
        if index >= self.number_of_accounts:
            new_index = index - self.number_of_accounts
            return self.generated_accounts[new_index]

        try:
            return self.provider.get_test_account(index)
        except (NotImplementedError, ProviderNotConnectedError):
            return self.generate_account(index=index)

    def generate_account(self, index: Optional[int] = None) -> "TestAccountAPI":
        new_index = (
            self.number_of_accounts + len(self.generated_accounts) if index is None else index
        )
        generated_account = generate_dev_accounts(
            self.mnemonic, 1, hd_path=self.hd_path, start_index=new_index
        )[0]
        account = self.init_test_account(
            new_index, generated_account.address, generated_account.private_key
        )

        # Only cache if being created outside the expected number of accounts.
        # Else, ends up cached twice and caused logic problems elsewhere.
        if new_index >= self.number_of_accounts:
            self.generated_accounts.append(account)

        return account

    @classmethod
    def init_test_account(cls, index: int, address: AddressType, private_key: str) -> "TestAccount":
        return TestAccount(
            index=index,
            address_str=address,
            private_key=private_key,
        )

    def reset(self):
        self.generated_accounts = []


class TestAccount(TestAccountAPI):
    index: int
    address_str: str
    private_key: str

    __test__ = False

    @property
    def alias(self) -> str:
        return f"TEST::{self.index}"

    @property
    def address(self) -> AddressType:
        return self.network_manager.ethereum.decode_address(self.address_str)

    @property
    def public_key(self) -> "HexBytes":
        return derive_public_key(HexBytes(self.private_key))

    @log_instead_of_fail(default="<TestAccount>")
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}_{self.index} {self.address_str}>"

    def sign_authorization(
        self,
        address: AddressType,
        chain_id: Optional[int] = None,
        nonce: Optional[int] = None,
    ) -> Optional[MessageSignature]:
        if chain_id is None:
            chain_id = self.provider.chain_id

        try:
            signed_authorization = EthAccount.sign_authorization(
                dict(
                    chainId=chain_id,
                    address=to_canonical_address(address),
                    nonce=nonce or self.nonce,
                ),
                self.private_key,
            )
        except AttributeError as e:
            # TODO: Remove `try..except` once web3 pinned `>=7`
            raise APINotImplementedError from e

        return MessageSignature(
            v=signed_authorization.y_parity,
            r=to_bytes(signed_authorization.r),
            s=to_bytes(signed_authorization.s),
        )

    def sign_message(self, msg: Any, **signer_options) -> Optional[MessageSignature]:
        # Convert str and int to SignableMessage if needed
        if isinstance(msg, str):
            msg = encode_defunct(text=msg)
        elif isinstance(msg, int):
            msg = to_hex(msg)
            msg = encode_defunct(hexstr=msg)
        elif isinstance(msg, EIP712Message):
            # Convert EIP712Message to SignableMessage for handling below
            msg = msg.signable_message

        # Process SignableMessage
        if isinstance(msg, SignableMessage):
            signed_msg = EthAccount.sign_message(msg, self.private_key)
            return MessageSignature(
                v=signed_msg.v,
                r=to_bytes(signed_msg.r),
                s=to_bytes(signed_msg.s),
            )
        return None

    def sign_transaction(
        self, txn: "TransactionAPI", **signer_options
    ) -> Optional["TransactionAPI"]:
        # Signs any transaction that's given to it.
        # NOTE: Using JSON mode, as only primitive types can be signed.
        tx_data = txn.model_dump(mode="json", by_alias=True, exclude={"sender"})
        private_key = PrivateKey(HexBytes(self.private_key))

        # NOTE: var name `sig_r` instead of `r` to avoid clashing with pdb commands.
        try:
            (
                sig_v,
                sig_r,
                sig_s,
                _,
            ) = sign_transaction_dict(private_key, tx_data)
        except TypeError as err:
            # Occurs when missing properties on the txn that are needed to sign.
            raise SignatureError(str(err), transaction=txn) from err

        # NOTE: Using `to_bytes(hexstr=to_hex(sig_r))` instead of `to_bytes(sig_r)` as
        #   a performance optimization.
        txn.signature = TransactionSignature(
            v=sig_v,
            r=to_bytes(hexstr=to_hex(sig_r)),
            s=to_bytes(hexstr=to_hex(sig_s)),
        )

        return txn

    def sign_raw_msghash(self, msghash: HexBytes) -> MessageSignature:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            signed_msg = sign_hash(msghash, self.private_key)

        return MessageSignature(
            v=signed_msg.v,
            r=to_bytes(signed_msg.r),
            s=to_bytes(signed_msg.s),
        )

    def set_delegate(self, contract: Union[BaseAddress, AddressType, str], **txn_kwargs):
        contract_address = self.conversion_manager.convert(contract, AddressType)
        sig = self.sign_authorization(contract_address, nonce=self.nonce + 1)
        auth = Authorization.from_signature(
            address=contract_address,
            chain_id=self.provider.chain_id,
            # NOTE: `tx` uses `self.nonce`
            nonce=self.nonce + 1,
            signature=sig,
        )
        tx = self.provider.network.ecosystem.create_transaction(
            type=TransactionType.SET_CODE,
            authorizations=[auth],
            sender=self,
            # NOTE: Cannot target `ZERO_ADDRESS`
            receiver=txn_kwargs.pop("receiver", None) or self,
            **txn_kwargs,
        )
        return self.call(tx)

    def remove_delegate(self, **txn_kwargs):
        sig = self.sign_authorization(ZERO_ADDRESS, nonce=self.nonce + 1)
        auth = Authorization.from_signature(
            chain_id=self.provider.chain_id,
            address=ZERO_ADDRESS,
            # NOTE: `tx` uses `self.nonce`
            nonce=self.nonce + 1,
            signature=sig,
        )
        tx = self.provider.network.ecosystem.create_transaction(
            type=TransactionType.SET_CODE,
            authorizations=[auth],
            sender=self,
            # NOTE: Cannot target `ZERO_ADDRESS`
            receiver=txn_kwargs.pop("receiver", None) or self,
            **txn_kwargs,
        )
        return self.call(tx)



from eth_account import Account as EthAccount

try:
    # Web3 v7
    from web3.middleware import ExtraDataToPOAMiddleware  # type: ignore
except ImportError:
    from web3.middleware import geth_poa_middleware as ExtraDataToPOAMiddleware  # type: ignore

try:
    from web3.providers import WebsocketProviderV2 as WebsocketProvider  # type: ignore
except ImportError:
    from web3.providers import WebSocketProvider as WebsocketProvider  # type: ignore


def sign_hash(msghash, private_key):
    try:
        # Web3 v7
        return EthAccount.unsafe_sign_hash(msghash, private_key)  # type: ignore
    except AttributeError:
        # Web3 v6
        return EthAccount.signHash(msghash, private_key)  # type: ignore


__all__ = [
    "ExtraDataToPOAMiddleware",
    "WebsocketProvider",
    "sign_hash",
]



import re
import shutil

import pytest

from ape.contracts import ContractContainer
from ape.utils import get_full_extension

from .utils import skip_projects, skip_projects_except

skip_non_compilable_projects = skip_projects(
    "bad-contracts",
    "empty-config",
    "no-config",
    "only-dependencies",
    "only-script-subdirs",
    "script",
    "test",
)


@skip_projects(
    "geth",
    "multiple-interfaces",
    "only-dependencies",
    "test",
    "bad-contracts",
    "script",
    "with-dependencies",
    "with-contracts",
)
def test_compile_missing_contracts_dir(ape_cli, runner, project):
    arg_lists = [["compile"], ["compile", "--include-dependencies"]]
    for arg_list in arg_lists:
        arg_list.extend(("--project", f"{project.path}"))
        result = runner.invoke(ape_cli, arg_list)
        assert result.exit_code == 0, result.output


@skip_non_compilable_projects
def test_compile(ape_cli, runner, integ_project, clean_cache):
    integ_project.clean()

    cmd = ("compile", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output

    # We have to load the manifest again because normally we are not referencing
    # the manifest in the same process as `ape compile`.
    integ_project.load_manifest()

    assert integ_project.manifest.contract_types

    # First time it compiles, it compiles the files with registered compilers successfully.
    # Files with multiple extensions are currently not supported.
    all_files = [f for f in integ_project.path.glob("contracts/**/*") if f.is_file()]

    # Don't expect directories that may happen to have `.json` in name
    # as well as hidden files, such as `.gitkeep`. Both examples are present
    # in the test project!
    excluded = (
        "Exclude.json",
        "IgnoreUsingRegex.json",
        "UnwantedContract.json",
        "tsconfig.json",
        "package.json",
        "package-lock.json",
    )
    expected_files = [
        f
        for f in all_files
        if f.name.count(".") == 1
        and f.is_file()
        and not f.name.startswith(".")
        and f.name not in excluded
        and get_full_extension(f) == ".json"
        and ".cache" not in [p.name for p in f.parents]
    ]
    unexpected_files = [f for f in all_files if f not in expected_files]

    # Extract manifest (note: same project is used in 2 instances here).
    manifest = integ_project.extract_manifest()

    non_json = [f for f in expected_files if get_full_extension(f) != ".json"]
    if len(non_json) > 0:
        assert manifest.compilers
    for file in expected_files:
        assert f"{integ_project.contracts_folder.name}/{file.name}" in manifest.sources

    missing = [f.name for f in expected_files if f.stem not in result.output]

    assert not missing, f"Missing: {', '.join(missing)}"

    for file in unexpected_files:
        assert file.stem not in result.output, f"Shouldn't have compiled {file.name}"

    # Copy in .build to show that those file won't compile.
    # (Anything in a .build is ignored, even if in a contracts folder to prevent accidents).
    build_path = integ_project.path / ".build"
    if integ_project.path != integ_project.contracts_folder:
        shutil.copytree(build_path, integ_project.contracts_folder / ".build")

    assert build_path.is_dir()
    try:
        result = runner.invoke(ape_cli, cmd, catch_exceptions=False)
        assert result.exit_code == 0, result.output
        assert "__local__.json" not in result.output
        # First time it compiles, it caches
        for file in integ_project.path.glob("contracts/**/*"):
            if file.is_file():
                assert file.name not in result.output

    finally:
        shutil.rmtree(integ_project.contracts_folder / ".build", ignore_errors=True)


@skip_projects_except("multiple-interfaces")
def test_compile_when_sources_change(ape_cli, runner, integ_project, clean_cache):
    source_path = integ_project.contracts_folder / "Interface.json"
    content = source_path.read_text()
    assert "bar" in content, "Test setup failed - unexpected content"

    result = runner.invoke(
        ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" in result.output
    assert "SUCCESS" in result.output
    assert "'local project' compiled." in result.output

    # Change the contents of a file.
    source_path = integ_project.contracts_folder / "Interface.json"
    modified_source_text = source_path.read_text().replace("bar", "foo")
    source_path.unlink()
    source_path.touch()
    source_path.write_text(modified_source_text, encoding="utf8")
    result = runner.invoke(
        ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" in result.output
    assert "SUCCESS" in result.output
    assert "'local project' compiled." in result.output

    # Verify that the next time, it does not need to recompile (no changes)
    result = runner.invoke(
        ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" not in result.output


@skip_projects_except("multiple-interfaces")
def test_compile_when_sources_change_problematically(ape_cli, runner, integ_project, clean_cache):
    """
    There was a bug when sources changes but had errors, that the old sources continued
    to be used and the errors were swallowed.
    """
    source_path = integ_project.contracts_folder / "Interface.json"
    content = source_path.read_text()
    assert "bar" in content, "Test setup failed - unexpected content"

    result = runner.invoke(
        ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
    )
    assert result.exit_code == 0, result.output

    # Change the contents of a file in a problematic way.
    source_path = integ_project.contracts_folder / "Interface.json"
    modified_source_text = source_path.read_text().replace("{", "BRACKET")
    source_path.unlink()
    source_path.touch()
    source_path.write_text(modified_source_text, encoding="utf8")
    result = runner.invoke(
        ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
    )
    assert result.exit_code != 0, result.output


@skip_projects_except("multiple-interfaces")
def test_compile_when_contract_type_collision(ape_cli, runner, integ_project, clean_cache):
    source_path = integ_project.contracts_folder / "Interface.json"
    temp_dir = integ_project.contracts_folder / "temp"

    def clean():
        if temp_dir.is_dir():
            shutil.rmtree(temp_dir)

    clean()
    source_copy = temp_dir / "Interface.json"
    expected = (
        r"ERROR: *\(CompilerError\) ContractType collision\. "
        r"Contracts '(.*\.json)' and '(.*\.json)' share the name 'Interface'\."
    )
    temp_dir.mkdir()
    try:
        source_copy.touch()
        source_copy.write_text(source_path.read_text(), encoding="utf8")
        result = runner.invoke(
            ape_cli, ("compile", "--project", f"{integ_project.path}"), catch_exceptions=False
        )
        assert result.exit_code == 1
        actual = result.output
        search_result = re.search(expected, actual)
        assert search_result, actual
        groups = search_result.groups()
        expected_group = {"contracts/Interface.json", "contracts/temp/Interface.json"}
        assert set(groups) == expected_group

    finally:
        clean()


@skip_projects_except("multiple-interfaces")
def test_compile_when_source_contains_return_characters(
    ape_cli, runner, integ_project, clean_cache
):
    """
    This tests a bugfix where a source file contained return-characters
    and that triggered endless re-compiles because it technically contains extra
    bytes than the ones that show up in the text.
    """
    source_path = integ_project.contracts_folder / "Interface.json"
    # Change the contents of a file to contain the '\r' character.
    modified_source_text = f"{source_path.read_text()}\r"
    source_path.unlink()
    source_path.touch()
    source_path.write_text(modified_source_text, encoding="utf8")
    arguments = ("compile", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" in result.output

    # Verify that the next time, it does not need to recompile (no changes)
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" not in result.output


@skip_projects_except("multiple-interfaces")
def test_can_access_contracts(integ_project, clean_cache):
    # This test does not use the CLI but still requires a project or run off of.
    assert integ_project.Interface, "Unable to access contract when needing to compile"
    assert integ_project.Interface, "Unable to access contract when not needing to compile"


@skip_projects_except("multiple-interfaces")
@pytest.mark.parametrize(
    "contract_path",
    ("Interface", "Interface.json", "contracts/Interface.json", "contracts/Interface"),
)
def test_compile_specified_contracts(clean_cache, ape_cli, runner, integ_project, contract_path):
    assert integ_project.path.is_dir(), "Setup failed - project missing."

    arguments = ("compile", contract_path, "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" in result.output

    # Already compiled.
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" not in result.output

    # Force recompile.
    result = runner.invoke(ape_cli, [*arguments, "--force"], catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "contracts/Interface.json" in result.output


@skip_projects_except("multiple-interfaces")
def test_compile_unknown_extension_does_not_compile(ape_cli, runner, integ_project, clean_cache):
    arguments = ("compile", "Interface.js", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 2, result.output
    assert "Error" in result.output
    assert "Source file 'Interface.js' not found." in result.output


@skip_projects_except("with-dependencies")
@pytest.mark.parametrize(
    "contract_path",
    (None, "contracts/", "Project", "contracts/Project.json"),
)
def test_compile_with_dependency(ape_cli, runner, integ_project, contract_path):
    cmd = ["compile", "--force", "--project", f"{integ_project.path}"]

    if contract_path:
        cmd.append(contract_path)

    result = runner.invoke(ape_cli, cmd, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    for name in (
        "default",
        "renamed-contracts-folder",
        "containing-sub-dependencies",
        "renamed-complex-contracts-folder",
        "renamed-contracts-folder-specified-in-config",
        "manifest-dependency",
    ):
        assert name in list(integ_project.dependencies.keys())
        dependency = integ_project.dependencies[name]["local"]
        assert isinstance(dependency[name], ContractContainer)


@skip_projects_except("with-dependencies")
def test_compile_individual_contract_excludes_other_contract(ape_cli, runner, integ_project):
    arguments = ("compile", "Project", "--force", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert "Other" not in result.output


@skip_projects_except("with-dependencies")
def test_compile_non_ape_project_deletes_ape_config_file(ape_cli, runner, integ_project):
    ape_config = integ_project.path / "default" / "ape-config.yaml"
    if ape_config.is_file():
        # Corrupted from a previous test.
        ape_config.unlink()

    arguments = ("compile", "Project", "--force", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output


@skip_projects_except("only-dependencies")
def test_compile_only_dependency(ape_cli, runner, integ_project, clean_cache, ape_caplog):
    expected_log_message = "Compiling sources/DependencyInProjectOnly.json"

    # Compile w/o --include-dependencies flag (nothing happens, but it doesn't fail).
    arguments: tuple[str, ...] = ("compile", "--force", "--project", f"{integ_project.path}")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 0, result.output
    assert expected_log_message not in result.output

    # Now, actually compile (using --include-dependencies)
    arguments = (
        "compile",
        "--force",
        "--project",
        f"{integ_project.path}",
        "--include-dependencies",
    )
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert expected_log_message in result.output

    # It should not need to compile again (no force).
    arguments = ("compile", "--project", f"{integ_project.path}", "--include-dependencies")
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert expected_log_message not in result.output

    # Ensure config reading works. Note: have to edit file here
    # because in-memory config updates only works when on the subprocess,
    # and the CLI has to reload the project.
    config_path = integ_project.path / "ape-config.yaml"
    integ_project.config.compile.include_dependencies = True
    integ_project.config.write_to_disk(config_path, replace=True)

    arguments = (
        "compile",
        "--force",
        "--project",
        f"{integ_project.path}",
        "--include-dependencies",
    )
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert result.exit_code == 1, result.output  # exit_code=1 because 1 dependency is bad.
    assert expected_log_message in result.output

    error_str = "Dependency 'dependency-that-is-not-installed' not installed."
    assert error_str in result.output


@skip_projects_except("with-contracts")
def test_raw_compiler_output_bytecode(integ_project):
    assert integ_project.RawVyperOutput.contract_type.runtime_bytecode.bytecode
    assert integ_project.RawSolidityOutput.contract_type.deployment_bytecode.bytecode


@skip_projects_except("with-contracts")
def test_compile_exclude(ape_cli, runner):
    result = runner.invoke(ape_cli, ("compile", "--force"), catch_exceptions=False)
    assert "Compiling 'contracts/Exclude.json'" not in result.output
    assert "Compiling 'contracts/IgnoreUsingRegex.json'" not in result.output
    assert "Compiling 'contracts/exclude_dir/UnwantedContract.json'" not in result.output


@skip_projects_except("with-contracts")
def test_compile_config_override(ape_cli, runner):
    arguments = (
        "compile",
        "--force",
        "--config-override",
        '{"compile": {"exclude": ["*ContractA*"]}}',
    )
    result = runner.invoke(ape_cli, arguments, catch_exceptions=False)
    assert "Compiling 'contracts/ContractA.json'" not in result.output



from abc import abstractmethod
from typing import TYPE_CHECKING, Optional

from ape.api.networks import NetworkAPI
from ape.utils.basemodel import BaseInterfaceModel

if TYPE_CHECKING:
    from ethpm_types import ContractType

    from ape.types.address import AddressType


class ExplorerAPI(BaseInterfaceModel):
    """
    An API class representing a blockchain explorer for a particular network
    in a particular ecosystem.
    """

    name: str  # Plugin name
    network: NetworkAPI

    @abstractmethod
    def get_address_url(self, address: "AddressType") -> str:
        """
        Get an address URL, such as for a transaction.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address.

        Returns:
            str: The URL.
        """

    @abstractmethod
    def get_transaction_url(self, transaction_hash: str) -> str:
        """
        Get the transaction URL for the given transaction.

        Args:
            transaction_hash (str): The transaction hash.

        Returns:
            str: The URL.
        """

    @abstractmethod
    def get_contract_type(self, address: "AddressType") -> Optional["ContractType"]:
        """
        Get the contract type for a given address if it has been published to this explorer.

        Args:
            address (:class:`~ape.types.address.AddressType`): The contract address.

        Returns:
            Optional[``ContractType``]: If not published, returns ``None``.
        """

    @abstractmethod
    def publish_contract(self, address: "AddressType"):
        """
        Publish a contract to the explorer.

        Args:
            address (:class:`~ape.types.address.AddressType`): The address of the deployed contract.
        """

    @classmethod
    def supports_chain(cls, chain_id: int) -> bool:
        """
        Returns ``True`` when the given chain ID is claimed to be
        supported by this explorer. Adhoc / custom networks rely on
        this feature to have automatic-explorer support. Explorer
        plugins should override this.

        Args:
            chain_id (int): The chain ID to check.

        Returns:
            bool
        """
        return False



import difflib
import sys
import tempfile
import time
import traceback
from collections.abc import Collection, Iterable
from functools import cached_property
from inspect import getframeinfo, stack
from pathlib import Path
from types import CodeType, TracebackType
from typing import TYPE_CHECKING, Any, Callable, Optional, Union

import click
from rich import print as rich_print

from ape.logging import LogLevel, logger

if TYPE_CHECKING:
    from eth_typing import HexStr
    from ethpm_types.abi import ConstructorABI, ErrorABI, MethodABI
    from ethpm_types.contract_type import ContractType

    from ape.api.networks import NetworkAPI
    from ape.api.providers import SubprocessProvider
    from ape.api.trace import TraceAPI
    from ape.api.transactions import ReceiptAPI, TransactionAPI
    from ape.managers.project import ProjectManager
    from ape.types.address import AddressType
    from ape.types.trace import SourceTraceback
    from ape.types.vm import BlockID, SnapshotID


FailedTxn = Union["TransactionAPI", "ReceiptAPI"]


class ApeException(Exception):
    """
    An exception raised by ape.
    """


class ApeIndexError(ApeException, IndexError):
    """
    An exception that is also an IndexError.
    Useful for nicely displaying IndexErrors.
    """


class APINotImplementedError(ApeException, NotImplementedError):
    """
    An error raised when an API class does not implement an abstract method.
    """


class AccountsError(ApeException):
    """
    Raised when a problem occurs when using accounts.
    """


class AliasAlreadyInUseError(AccountsError):
    """
    Raised when attempting to add an account using an alias
    that already maps to another account.
    """

    def __init__(self, alias: str):
        self.alias = alias
        super().__init__(f"Account with alias '{alias}' already in use.")


class SignatureError(AccountsError):
    """
    Raised when there are issues with signing.
    """

    def __init__(self, message: str, transaction: Optional["TransactionAPI"] = None):
        self.transaction = transaction
        super().__init__(message)


class ContractDataError(ApeException):
    """
    Raised when issues occur with local contract.
    **NOTE**: This error has nothing to do with on-chain
    contract logic errors; it is more about ABI-related
    issues and alike.
    """


class MissingDeploymentBytecodeError(ContractDataError):
    """
    Raised when trying to deploy an interface or empty data.
    """

    def __init__(self, contract_type: "ContractType"):
        message = "Cannot deploy: contract"
        if name := contract_type.name:
            message = f"{message} '{name}'"

        message = (
            f"{message} has no deployment-bytecode. Are you attempting to deploy an interface?"
        )
        super().__init__(message)


class ArgumentsLengthError(ContractDataError):
    """
    Raised when calling a contract method with the wrong number of arguments.
    """

    def __init__(
        self,
        arguments_length: int,
        inputs: Union["MethodABI", "ConstructorABI", int, list, None] = None,
        **kwargs,
    ):
        prefix = (
            f"The number of the given arguments ({arguments_length}) "
            f"do not match what is defined in the ABI"
        )
        if inputs is None:
            super().__init__(f"{prefix}.")
            return

        inputs_ls: list[Union[MethodABI, ConstructorABI, int]] = (
            inputs if isinstance(inputs, list) else [inputs]
        )
        if not inputs_ls:
            suffix = ""
        elif any(not isinstance(x, int) for x in inputs_ls):
            # Handle ABI arguments
            parts = ""
            for idx, ipt in enumerate(inputs_ls):
                if isinstance(ipt, int):
                    part = f"{ipt}"
                else:
                    # Signature without outputs.
                    input_args = ", ".join(i.signature for i in ipt.inputs)
                    part = f"{getattr(ipt, 'name', '__init__')}({input_args})"

                parts = f"{parts}\n\t{part}"

            suffix = f":\n{parts}"

        else:
            # Was only given integers.
            options = ", ".join([str(x) for x in inputs_ls])
            one_of = "one of " if len(inputs_ls) > 1 else ""
            suffix = f" ({one_of}{options})"

        super().__init__(f"{prefix}{suffix}")


class DecodingError(ContractDataError):
    """
    Raised when issues occur while decoding data from
    a contract call, transaction, or event.
    """

    def __init__(self, message: Optional[str] = None):
        message = message or "Output corrupted."
        super().__init__(message)


class MethodNonPayableError(ContractDataError):
    """
    Raises when sending funds to a non-payable method
    """


_TRACE_ARG = Optional[Union["TraceAPI", Callable[[], Optional["TraceAPI"]]]]
_SOURCE_TRACEBACK_ARG = Optional[
    Union["SourceTraceback", Callable[[], Optional["SourceTraceback"]]]
]


class TransactionError(ApeException):
    """
    Raised when issues occur related to transactions.
    """

    DEFAULT_MESSAGE = "Transaction failed."

    def __init__(
        self,
        message: Optional[str] = None,
        base_err: Optional[Exception] = None,
        code: Optional[int] = None,
        txn: Optional[FailedTxn] = None,
        trace: _TRACE_ARG = None,
        contract_address: Optional["AddressType"] = None,
        source_traceback: _SOURCE_TRACEBACK_ARG = None,
        project: Optional["ProjectManager"] = None,
        set_ape_traceback: bool = False,  # Overridden in ContractLogicError
    ):
        message = message or (str(base_err) if base_err else self.DEFAULT_MESSAGE)
        self.message = message
        self.base_err = base_err
        self.code = code
        self.txn = txn
        self._trace = trace
        self.contract_address = contract_address
        self._source_traceback = source_traceback
        self._project = project
        ex_message = f"({code}) {message}" if code else message

        # Finalizes expected revert message.
        super().__init__(ex_message)

        self._attempted_source_traceback = False
        if set_ape_traceback:
            self.with_ape_traceback()

    @property
    def address(self) -> Optional["AddressType"]:
        if addr := self.contract_address:
            return addr

        receiver = getattr(self.txn, "receiver", None)
        if receiver in (None, "0x0000000000000000000000000000000000000000"):
            # Check if deploy
            if addr := getattr(self.txn, "contract_address", None):
                return addr

        return receiver

    @cached_property
    def contract_type(self) -> Optional["ContractType"]:
        if not (address := self.address):
            # Contract address not found.
            return None

        # Lazy import because of exceptions.py root nature.
        from ape.utils.basemodel import ManagerAccessMixin

        try:
            return ManagerAccessMixin.chain_manager.contracts.get(address)
        except (RecursionError, ProviderNotConnectedError):
            return None

    @property
    def trace(self) -> Optional["TraceAPI"]:
        tr = self._trace
        if callable(tr):
            result = tr()
            self._trace = result
            return result

        return tr

    @trace.setter
    def trace(self, value):
        self._trace = value

    @property
    def source_traceback(self) -> Optional["SourceTraceback"]:
        tb = self._source_traceback
        result: Optional[SourceTraceback]
        if not self._attempted_source_traceback and tb is None and self.txn is not None:
            result = _get_ape_traceback_from_tx(self.txn)
            # Prevent re-trying.
            self._attempted_source_traceback = True
        elif callable(tb):
            result = tb()
        else:
            result = tb

        self._source_traceback = result
        return result

    @source_traceback.setter
    def source_traceback(self, value):
        self._source_traceback = value

    def _get_ape_traceback(self) -> Optional[TracebackType]:
        if src_tb := self.source_traceback:
            # Create a custom Pythonic traceback using lines from the sources
            # found from analyzing the trace of the transaction.
            if py_tb := _get_custom_python_traceback(self, src_tb, project=self._project):
                return py_tb

        return None

    def with_ape_traceback(self):
        return self.with_traceback(self._get_ape_traceback())


class VirtualMachineError(TransactionError):
    """
    Raised when a transaction error occurs in a virtual machine.
    """


class ContractLogicError(VirtualMachineError):
    """
    Raised when there is a contract-defined revert,
    such as from an assert/require statement.
    """

    def __init__(
        self,
        revert_message: Optional[str] = None,
        txn: Optional[FailedTxn] = None,
        trace: _TRACE_ARG = None,
        contract_address: Optional["AddressType"] = None,
        source_traceback: _SOURCE_TRACEBACK_ARG = None,
        base_err: Optional[Exception] = None,
        project: Optional["ProjectManager"] = None,
        set_ape_traceback: bool = True,  # Overridden default.
    ):
        self.txn = txn
        self.contract_address = contract_address
        super().__init__(
            base_err=base_err,
            contract_address=contract_address,
            message=revert_message,
            project=project,
            set_ape_traceback=set_ape_traceback,
            source_traceback=source_traceback,
            trace=trace,
            txn=txn,
        )
        if not revert_message and source_traceback is not None and (dev := self.dev_message):
            try:
                # Attempt to use dev message as main exception message.
                self.message = dev
            except Exception:
                pass

    @property
    def revert_message(self):
        return self.message

    @property
    def dev_message(self) -> Optional[str]:
        """
        The dev-string message of the exception.

        Raises:
            ``ValueError``: When unable to get dev message.
        """
        return self.source_traceback.revert_type if self.source_traceback else None

    @classmethod
    def from_error(cls, err: Exception):
        """
        Creates this class from the error message of the given
        error.

        This should be overridden whenever possible to handle
        provider-specific use-cases for raising this error.
        """
        return cls(str(err))


class OutOfGasError(VirtualMachineError):
    """
    Raised when detecting a transaction failed because it ran
    out of gas.
    """

    def __init__(
        self,
        code: Optional[int] = None,
        txn: Optional[FailedTxn] = None,
        base_err: Optional[Exception] = None,
        set_ape_traceback: bool = False,
    ):
        super().__init__(
            "The transaction ran out of gas.",
            code=code,
            txn=txn,
            base_err=base_err,
            set_ape_traceback=set_ape_traceback,
        )


class NetworkError(ApeException):
    """
    Raised when a problem occurs when using blockchain networks.
    """


class EcosystemNotFoundError(NetworkError):
    """
    Raised when the ecosystem with the given name was not found.
    """

    def __init__(self, ecosystem: str, options: Optional[Collection[str]] = None):
        self.ecosystem = ecosystem
        self.options = options
        message = f"No ecosystem named '{ecosystem}'."
        if options:
            close_matches = difflib.get_close_matches(ecosystem, options, cutoff=0.6)
            if close_matches:
                message = f"{message} Did you mean '{', '.join(close_matches)}'?"
            else:
                # No close matches. Show all the options.
                options_str = "\n".join(sorted(options))
                message = f"{message} Options:\n{options_str}"

        super().__init__(message)


class NetworkNotFoundError(NetworkError):
    """
    Raised when the network with the given name was not found.
    """

    def __init__(
        self,
        network: str,
        ecosystem: Optional[str] = None,
        options: Optional[Collection[str]] = None,
    ):
        self.network = network
        options = options or []
        if network in options:
            # Only seen in testing scenarios. Not realistic.
            raise ValueError(
                f"{network} found in options. Should not have gotten `NetworkNotFoundError`."
            )

        if options:
            message = (
                f"No network in '{ecosystem}' named '{network}'."
                if ecosystem
                else f"No network named '{network}'."
            )
            close_matches = difflib.get_close_matches(network, options, cutoff=0.6)
            if close_matches:
                message = f"{message} Did you mean '{', '.join(close_matches)}'?"
            else:
                # No close matches - show all options.
                options_str = "\n".join(sorted(options))
                message = f"{message} Options:\n{options_str}"

        elif ecosystem:
            message = f"'{ecosystem}' has no networks."
        else:
            message = "No networks found."

        super().__init__(message)


class ProviderNotFoundError(NetworkError):
    """
    Raised when the provider with the given name was not found.
    """

    def __init__(
        self,
        provider: str,
        network: Optional[str] = None,
        ecosystem: Optional[str] = None,
        options: Optional[Collection[str]] = None,
    ):
        self.provider = provider
        self.network = network
        self.ecosystem = ecosystem
        message = f"No provider named '{provider}'"
        if network:
            message = f"{message} in network '{network}'"
        if ecosystem:
            message = f"{message} in ecosystem '{ecosystem}'"
        if options:
            close_matches = difflib.get_close_matches(provider, options, cutoff=0.6)
            if close_matches:
                message = f"{message}. Did you mean '{', '.join(close_matches)}'?"
            else:
                # No close matches. Show all provider options.
                options_str = "\n".join(sorted(options))
                message = f"{message}. Options:\n{options_str}"
        else:
            message = f"{message}."

        super().__init__(message)


class CompilerError(ApeException):
    """
    Raised when unable to compile.
    """


class ProjectError(ApeException):
    """
    Raised when problems occur in a project.
    """


class ApeAttributeError(ProjectError, AttributeError):
    """
    Raised when trying to access items via ``.`` access.
    """

    def __init__(self, msg: str, base_err: Optional[Exception] = None):
        self.base_err = base_err
        super().__init__(msg)


class UnknownVersionError(ProjectError):
    """
    Raised when trying to install an unknown version of a package.
    """

    def __init__(self, version: str, name: str):
        super().__init__(f"Unknown version '{version}' for repo '{name}'.")


class ConversionError(ApeException):
    """
    Raised when unable to convert a value.
    """


class ProviderError(ApeException):
    """
    Raised when a problem occurs when using providers.
    """


class BlockNotFoundError(ProviderError):
    """
    Raised when unable to find a block.
    """

    def __init__(self, block_id: "BlockID", reason: Optional[str] = None):
        if isinstance(block_id, bytes):
            block_id_str = block_id.hex()
            if not block_id_str.startswith("0x"):
                block_id_str = f"0x{block_id_str}"
        else:
            block_id_str: HexStr = f"{block_id}"  # type: ignore

        message = (
            "Missing latest block."
            if block_id == "latest"
            else f"Block with ID '{block_id_str}' not found."
        )
        if reason:
            message = f"{message} Reason: {reason}"

        super().__init__(message)


class TransactionNotFoundError(ProviderError):
    """
    Raised when unable to find a transaction.
    """

    def __init__(self, transaction_hash: Optional[str] = None, error_message: Optional[str] = None):
        message = (
            f"Transaction '{transaction_hash}' not found."
            if transaction_hash
            else "Transaction not found"
        )
        suffix = f" Error: {error_message}" if error_message else ""
        super().__init__(f"{message}{suffix}")


class NetworkMismatchError(ProviderError):
    """
    Raised when connecting a provider to the wrong network.
    """

    def __init__(self, chain_id: int, network: "NetworkAPI"):
        message = (
            f"Provider connected to chain ID '{chain_id}', which does not match "
            f"network chain ID '{network.chain_id}'. Are you connected to '{network.name}'?"
        )
        super().__init__(message)


class ProviderNotConnectedError(ProviderError):
    """
    Raised when not connected to a provider.
    """

    def __init__(self):
        super().__init__("Not connected to a network provider.")


class ConfigError(ApeException):
    """
    Raised when a problem occurs from the configuration file.
    """


class ChainError(ApeException):
    """
    Raised when problems occur in the :class:`~ape.managers.chain.ChainManager`.
    """


class ContractNotFoundError(ChainError):
    """
    Raised when a contract is not found at an address.
    """

    # TODO: In 0.9, pass in provider object directly (instead of network choice + name)
    def __init__(self, address: "AddressType", has_explorer: bool, network_choice: str):
        msg = f"Failed to get contract type for address '{address}'."

        # NOTE: Network name is optional to avoid breaking change.
        choice_parts = network_choice.split(":")
        if len(choice_parts) > 1:
            network_name = network_choice.split(":")[1]
        else:
            network_name = network_choice

        if has_explorer:
            msg += " Contract may need verification."
        elif network_name != "local":
            # Only bother mentioning explorer plugins if we are not the local network.
            msg += (
                f" Current network '{network_choice}' has no associated "
                "explorer plugin. Try installing an explorer plugin using "
                f"{click.style(text='ape plugins install etherscan', fg='green')}, "
                "or using a network with explorer support."
            )

        super().__init__(msg)


class UnknownSnapshotError(ChainError):
    """
    Raised when given an unknown snapshot ID.
    """

    def __init__(self, snapshot_id: "SnapshotID"):
        snapshot_id_str: str
        if isinstance(snapshot_id, bytes):
            # Is block hash. Logic borrowed from `eth_utils.humanize_hash()`.
            if len(snapshot_id) <= 5:
                snapshot_id_str = snapshot_id.hex()
            else:
                value_hex = snapshot_id.hex()
                head = value_hex[:4]
                tail_idx = -1 * 4
                tail = value_hex[tail_idx:]
                snapshot_id_str = f"{head}..{tail}"

            snapshot_id_str = (
                f"0x{snapshot_id_str}" if not snapshot_id_str.startswith("0x") else snapshot_id_str
            )

        else:
            snapshot_id_str = f"{snapshot_id}"  # type: ignore

        super().__init__(f"Unknown snapshot ID '{snapshot_id_str}'.")


class QueryEngineError(ApeException):
    """
    Raised when issues occur in a query engine.
    """


class SubprocessError(ApeException):
    """
    An error raised whilst managing a subprocess.
    """


class SubprocessTimeoutError(SubprocessError):
    """
    A context-manager exception that raises if its operations exceed
    the given timeout seconds.

    This implementation was inspired from py-geth.
    """

    def __init__(
        self,
        provider: "SubprocessProvider",
        message: Optional[str] = None,
        seconds: Optional[int] = None,
        exception: Optional[Exception] = None,
        *args,
        **kwargs,
    ):
        self._provider = provider
        self._message = message or "Timed out waiting for process."
        self._seconds = seconds
        self._exception = exception
        self._start_time: Optional[float] = None
        self._is_running: Optional[bool] = None

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return False

    def __str__(self) -> str:
        if self._seconds in [None, ""]:
            return ""

        return self._message

    @property
    def expire_at(self) -> float:
        if self._seconds is None:
            raise ValueError("Timeouts with 'seconds == None' do not have an expiration time.")
        elif self._start_time is None:
            raise ValueError("Timeout has not been started.")

        return self._start_time + self._seconds

    def start(self):
        if self._is_running is not None:
            raise ValueError("Timeout has already been started.")

        self._start_time = time.time()
        self._is_running = True

    def check(self):
        if self._is_running is None:
            raise ValueError("Timeout has not been started.")

        elif self._is_running is False:
            raise ValueError("Timeout has already been cancelled.")

        elif self._seconds is None:
            return

        elif time.time() > self.expire_at:
            self.cancel()
            self._is_running = False

            if isinstance(self._exception, type):
                raise self._exception(str(self))

            elif isinstance(self._exception, Exception):
                raise self._exception

            raise self

    def cancel(self):
        if self._provider:
            self._provider.stop()

        self._is_running = False


class RPCTimeoutError(SubprocessTimeoutError):
    def __init__(
        self,
        provider: "SubprocessProvider",
        seconds: Optional[int] = None,
        exception: Optional[Exception] = None,
        *args,
        **kwargs,
    ):
        error_message = (
            "Timed out waiting for successful RPC connection to "
            f"the '{provider.process_name}' process ({seconds} seconds)."
        )
        kwargs["message"] = error_message
        if seconds:
            kwargs["seconds"] = seconds
        if exception:
            kwargs["exception"] = exception

        super().__init__(provider, *args, **kwargs)


class PluginInstallError(ApeException):
    """
    An error to use when installing a plugin fails.
    """


class PluginVersionError(PluginInstallError):
    """
    An error related to specified plugin version.
    """

    def __init__(
        self, operation: str, reason: Optional[str] = None, resolution: Optional[str] = None
    ):
        message = f"Unable to {operation} plugin."
        if reason:
            message = f"{message}\nReason: {reason}"
        if resolution:
            message = f"{message}\nTo resolve: {resolution}"

        super().__init__(message)


def handle_ape_exception(err: ApeException, base_paths: Iterable[Union[Path, str]]) -> bool:
    """
    Handle a transaction error by showing relevant stack frames,
    including custom contract frames added to the exception.
    This method must be called within an ``except`` block or with
    an exception on the exc-stack.

    Args:
        err (:class:`~ape.exceptions.TransactionError`): The transaction error
          being handled.
        base_paths (Optional[Iterable[Union[Path, str]]]): Optionally include additional
          source-path prefixes to use when finding relevant frames.

    Returns:
        bool: ``True`` if outputted something.
    """
    home_str = str(Path.home())
    if not (relevant_frames := _get_relevant_frames(base_paths)):
        return False

    formatted_tb = [x.replace(home_str, "$HOME") for x in relevant_frames]
    rich_print(f"\n{''.join(formatted_tb)}")

    # Prevent double logging of a traceback by using `show_traceback=False`.
    logger.error(Abort.from_ape_exception(err, show_traceback=False))
    return True


def _get_relevant_frames(base_paths: Iterable[Union[Path, str]]):
    # Abstracted for testing easement.
    tb = traceback.extract_tb(sys.exc_info()[2])
    if relevant_tb := [f for f in tb if any(str(p) in f.filename for p in base_paths)]:
        return [x for x in traceback.format_list(relevant_tb)]

    return []


class Abort(click.ClickException):
    """
    A wrapper around a CLI exception. When you raise this error,
    the error is nicely printed to the terminal. This is
    useful for all user-facing errors.
    """

    def __init__(self, message: Optional[str] = None):
        if not message:
            caller = getframeinfo(stack()[1][0])
            file_path = Path(caller.filename)
            location = file_path.name if file_path.is_file() else caller.filename
            message = f"Operation aborted in {location}::{caller.function} on line {caller.lineno}."

        super().__init__(message)

    @classmethod
    def from_ape_exception(cls, exc: ApeException, show_traceback: Optional[bool] = None):
        show_traceback = (
            logger.level == LogLevel.DEBUG.value if show_traceback is None else show_traceback
        )
        if show_traceback:
            tb = traceback.format_exc()
            err_message = tb or str(exc)
        else:
            err_message = str(exc)

        err_type_name = getattr(type(exc), "__name__", "Exception")
        return Abort(f"({err_type_name}) {err_message}")

    def show(self, file=None):
        """
        Override default ``show`` to print CLI errors in red text.
        """

        logger.error(self.format_message())


class CustomError(ContractLogicError):
    """
    An error defined in a smart contract.
    """

    def __init__(
        self,
        abi: "ErrorABI",
        inputs: dict[str, Any],
        txn: Optional[FailedTxn] = None,
        trace: _TRACE_ARG = None,
        contract_address: Optional["AddressType"] = None,
        base_err: Optional[Exception] = None,
        source_traceback: _SOURCE_TRACEBACK_ARG = None,
    ):
        self.abi = abi
        self.inputs = inputs

        if inputs:
            message = ", ".join(sorted([f"{k}={v}" for k, v in inputs.items()]))
        else:
            # Name of the custom error is all custom info.
            message = TransactionError.DEFAULT_MESSAGE

        super().__init__(
            message,
            base_err=base_err,
            contract_address=contract_address,
            source_traceback=source_traceback,
            trace=trace,
            txn=txn,
        )

    @property
    def name(self) -> str:
        """
        The name of the error.
        """
        return self.abi.name

    def __repr__(self) -> str:
        name = self.__class__.__name__  # Custom error name
        calldata = ", ".join(sorted([f"{k}={v}" for k, v in self.inputs.items()])) or ""
        return f"{name}({calldata})"


def _get_ape_traceback_from_tx(txn: FailedTxn) -> Optional["SourceTraceback"]:
    from ape.api.transactions import ReceiptAPI

    try:
        receipt: ReceiptAPI = txn if isinstance(txn, ReceiptAPI) else txn.receipt  # type: ignore
    except Exception:
        # Receipt not real enough, maybe was a re-played call.
        return None

    if not receipt:
        return None

    try:
        ape_traceback = receipt.source_traceback
    except (ApeException, NotImplementedError):
        return None

    if ape_traceback is None or not len(ape_traceback):
        return None

    return ape_traceback


def _get_custom_python_traceback(
    err: TransactionError,
    ape_traceback: "SourceTraceback",
    project: Optional["ProjectManager"] = None,
) -> Optional[TracebackType]:
    # Manipulate python traceback to show lines from contract.
    # Help received from Jinja lib:
    #  https://github.com/pallets/jinja/blob/main/src/jinja2/debug.py#L142

    if project is None:
        from ape.utils.basemodel import ManagerAccessMixin as access

        project = access.local_project

    if not (base_path := getattr(project, "path", None)):
        # TODO: Add support for manifest-projects.
        return None

    _, exc_value, tb = sys.exc_info()
    depth = None
    idx = len(ape_traceback) - 1
    frames = []
    while tb is not None:
        if not tb.tb_frame.f_code.co_filename.startswith(str(base_path)):
            # Ignore frames outside the project.
            # This allows both contract code an scripts to appear.
            tb = tb.tb_next
            continue

        frames.append(tb)
        tb = tb.tb_next

    while (depth is None or depth > 1) and idx >= 0:
        exec_item = ape_traceback[idx]
        if depth is not None and exec_item.depth >= depth:
            # Wait for decreasing depth.
            idx -= 1
            continue

        depth = exec_item.depth

        # NOTE: Use the last lineno executed as "the line number".
        lineno = exec_item.begin_lineno if exec_item.end_lineno is None else exec_item.end_lineno
        if lineno is None:
            idx -= 1
            continue

        if exec_item.source_path is None:
            # File is not local. Create a temporary file in its place.
            # This is necessary for tracebacks to work in Python.
            temp_file = tempfile.NamedTemporaryFile(prefix="unknown_contract_")
            filename = temp_file.name
        else:
            filename = str(exec_item.source_path)

        # Raise an exception at the correct line number.
        py_code: CodeType = compile(
            "\n" * (lineno - 1) + "raise __ape_exception__", filename, "exec"
        )
        py_code = py_code.replace(co_name=exec_item.closure.name)

        # Execute the new code to get a new (fake) tb with contract source info.
        try:
            exec(py_code, {"__ape_exception__": err}, {})
        except BaseException:
            real_tb = sys.exc_info()[2]
            fake_tb = getattr(real_tb, "tb_next", None)
            if isinstance(fake_tb, TracebackType):
                frames.append(fake_tb)

        idx -= 1

    if not frames:
        return None

    tb_next = None
    for tb in frames:
        tb.tb_next = tb_next
        tb_next = tb

    return frames[-1]



import pytest
from eth_pydantic_types import HexBytes
from eth_utils import to_hex

from ape.exceptions import ConversionError
from ape.managers.converters import HexConverter, HexIntConverter, HexIterableConverter


@pytest.mark.parametrize("val", ("0xA100", "0x0A100", "0x00a100"))
def test_hex_str(convert, val):
    assert convert(val, int) == 0xA100
    assert int(to_hex(convert(val, bytes)), 16) == int(to_hex(0xA100), 16)


def test_int_str(convert):
    assert convert("100", int) == 100


def test_missing_prefix(convert):
    hex_value = "A100"

    assert not HexConverter().is_convertible(hex_value)
    assert not HexIntConverter().is_convertible(hex_value)

    with pytest.raises(ConversionError):
        convert(hex_value, int)


@pytest.mark.parametrize(
    "calldata,expected",
    (
        (
            ["0x123456", "0xabcd"],
            "0x123456abcd",
        ),
        (
            [HexBytes("0x123456"), "0xabcd"],
            "0x123456abcd",
        ),
        (
            ("0x123456", "0xabcd"),
            "0x123456abcd",
        ),
    ),
)
def test_hex_concat(calldata, expected, convert):
    assert HexIterableConverter().is_convertible(calldata)
    assert convert(calldata, bytes) == HexBytes(expected)



import pytest

from ape.utils.misc import LOCAL_NETWORK_NAME
from tests.conftest import geth_process_test


@pytest.fixture
def mock_geth_sepolia(ethereum, geth_provider, geth_contract):
    """
    Temporarily tricks Ape into thinking the local network
    is Sepolia so we can test features that require a live
    network.
    """
    # Ensuring contract exists before hack.
    # This allows the network to be past genesis which is more realistic.
    _ = geth_contract
    geth_provider.network.name = "sepolia"
    yield geth_provider.network
    geth_provider.network.name = LOCAL_NETWORK_NAME


@geth_process_test
def test_fork_upstream_provider(networks, mock_geth_sepolia, geth_provider, mock_fork_provider):
    uri = "http://example.com/node"
    orig = geth_provider.provider_settings.get("uri")
    geth_provider.provider_settings["uri"] = uri
    try:
        with networks.fork():
            call = mock_fork_provider.partial_call

        settings = call[1]["provider_settings"]["fork"]["ethereum"]["sepolia"]
        actual = settings["upstream_provider"]
        assert actual == uri
    finally:
        # Restore.
        if orig:
            geth_provider.provider_settings["uri"] = orig
        else:
            del geth_provider.provider_settings["uri"]


# NOTE: Test is flakey because random URLs may be offline when test runs; avoid CI failure.
@pytest.mark.flaky(reruns=5)
@geth_process_test
@pytest.mark.parametrize(
    "connection_str", ("moonbeam:moonriver", "https://moonriver.api.onfinality.io/public")
)
def test_parse_network_choice_evmchains(networks, connection_str):
    """
    Show we can (without having a plugin installed) connect to a network
    that evm-chains knows about.
    """
    with networks.parse_network_choice(connection_str) as moon_provider:
        assert moon_provider.network.name == "moonriver"
        assert moon_provider.network.ecosystem.name == "moonbeam"

        # When including the HTTP URL in the network choice,
        # `provider.network_choice` should also include it.
        # This ensures when relying on `provider.network_choice` for
        # multiple connections that they all use the same HTTP URI.
        expected_network_choice = (
            f"moonbeam:moonriver:{connection_str}"
            if "http" in connection_str
            else f"{connection_str}:node"
        )
        assert moon_provider.network_choice == expected_network_choice


@geth_process_test
def test_parse_network_choice_pid(geth_provider, networks):
    pid = None
    if proc := geth_provider.process:
        pid = proc.pid  # Still potentially None.

    if pid is None:
        pid = next(iter(networks.running_nodes.lookup_processes(geth_provider)), None)

    if pid is None:
        pytest.fail("Cannot find Geth process - might be running _before_ tests.")

    # Show we are able to connect to providers via PID URL.
    with networks.parse_network_choice(f"pid://{pid}") as provider:
        assert provider.is_connected
        assert provider.network_choice == f"ethereum:local:{geth_provider.ipc_path}"



import json
from pathlib import Path
from re import Pattern
from typing import cast

import pytest
from ethpm_types import ContractType, ErrorABI

from ape.contracts import ContractContainer
from ape.exceptions import APINotImplementedError, CompilerError, ContractLogicError, CustomError
from ape.types.address import AddressType
from ape_compile.config import Config


def test_get_imports(project, compilers):
    # See ape-solidity for better tests
    assert not compilers.get_imports(project.sources.paths)


def test_get_compiler(compilers):
    compiler = compilers.get_compiler("ethpm")
    assert compiler.name == "ethpm"
    assert compilers.get_compiler("foobar") is None


def test_get_compiler_with_settings(compilers, mock_compiler, project_with_contract):
    _ = compilers.registered_compilers  # Ensures cached property is set.

    # Hack in our mock compiler.
    compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler

    try:
        compiler_0 = compilers.get_compiler("mock", settings={"bar": "foo"})
        compiler_1 = compiler_0.get_compiler("mock", settings={"foo": "bar"})

    finally:
        if mock_compiler.ext in compilers.__dict__.get("registered_compilers", {}):
            del compilers.__dict__["registered_compilers"][mock_compiler.ext]

    assert compiler_0.compiler_settings != compiler_1.compiler_settings
    assert id(compiler_0) != id(compiler_1)


def test_getattr(compilers):
    compiler = compilers.ethpm
    assert compiler.name == "ethpm"
    expected = (
        r"'CompilerManager' object has no attribute 'foobar'\. "
        r"Also checked extra\(s\) 'compilers'\."
    )

    with pytest.raises(AttributeError, match=expected):
        _ = compilers.foobar


def test_supports_tracing(compilers):
    assert not compilers.ethpm.supports_source_tracing


def test_flatten_contract(compilers, project_with_contract):
    """
    Positive tests exist in compiler plugins that implement this behavior.b
    """
    source_id = project_with_contract.ApeContract0.contract_type.source_id
    path = project_with_contract.contracts_folder / source_id

    with pytest.raises(APINotImplementedError):
        compilers.flatten_contract(path)

    expected = r"Unable to flatten contract\. Missing compiler for '.foo'\."
    with pytest.raises(CompilerError, match=expected):
        compilers.flatten_contract(Path("contract.foo"))


@pytest.mark.parametrize("factory", (str, Path))
def test_compile(compilers, project_with_contract, factory):
    """
    Testing both stringified paths and path-object paths.
    """
    path = next(iter(project_with_contract.sources.paths))
    actual = compilers.compile((factory(path),), project=project_with_contract)
    contract_name = path.stem
    assert contract_name in [x.name for x in actual]


def test_compile_contract_type_collision(compilers, project_with_contract, mock_compiler):
    _ = compilers.registered_compilers  # Ensures cached property is set.

    # Hack in our mock compiler.
    compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler

    try:
        # Make contracts of type .__mock__ with the same names.
        existing_contract = next(iter(project_with_contract.contracts.values()))
        assert existing_contract.source_id, "Setup failed: Contract missing source ID!"
        existing_path = project_with_contract.path / existing_contract.source_id
        new_contract = project_with_contract.path / existing_contract.source_id.replace(
            ".json", mock_compiler.ext
        )
        new_contract.write_text("foobar", encoding="utf8")
        to_compile = [existing_path, new_contract]
        compile = compilers.compile(to_compile, project=project_with_contract)

        with pytest.raises(CompilerError, match="ContractType collision.*"):
            # Must include existing contract in case not yet compiled.
            _ = list(compile)

    finally:
        if mock_compiler.ext in compilers.__dict__.get("registered_compilers", {}):
            del compilers.__dict__["registered_compilers"][mock_compiler.ext]


def test_compile_empty(compilers):
    # Also, we are asserting it does no fail.
    assert list(compilers.compile([])) == []


def test_compile_with_settings(mock_compiler, compilers, project_with_contract):
    new_contract = project_with_contract.path / f"AMockContract{mock_compiler.ext}"
    new_contract.write_text("foobar", encoding="utf8")
    settings = {"mock": {"foo": "bar"}}
    _ = compilers.registered_compilers  # Ensures cached property is set.
    # Hack in our mock compiler.
    compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler

    try:
        list(compilers.compile([new_contract], project=project_with_contract, settings=settings))

    finally:
        if mock_compiler.ext in compilers.__dict__.get("registered_compilers", {}):
            del compilers.__dict__["registered_compilers"][mock_compiler.ext]

    actual = mock_compiler.method_calls[0][2]["settings"]
    assert actual == settings["mock"]


def test_compile_errors(mock_compiler, compilers, project_with_contract):
    new_contract = project_with_contract.path / f"AMockContract{mock_compiler.ext}"
    new_contract.write_text("foobar", encoding="utf8")

    class MyCustomCompilerError(CompilerError):
        pass

    mock_compiler.compile.side_effect = MyCustomCompilerError
    _ = compilers.registered_compilers  # Ensures cached property is set.
    # Hack in our mock compiler.
    compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler

    try:
        with pytest.raises(MyCustomCompilerError):
            list(compilers.compile([new_contract], project=project_with_contract))

    finally:
        if mock_compiler.ext in compilers.__dict__.get("registered_compilers", {}):
            del compilers.__dict__["registered_compilers"][mock_compiler.ext]


def test_compile_multiple_errors(
    mock_compiler, make_mock_compiler, compilers, project_with_contract
):
    """
    Simulating getting errors from multiple compilers.
    We should get all the errors.
    """
    second_mock_compiler = make_mock_compiler("mock2")
    new_contract_0 = project_with_contract.path / f"AMockContract{mock_compiler.ext}"
    new_contract_0.write_text("foobar", encoding="utf8")
    new_contract_1 = project_with_contract.path / f"AMockContract{second_mock_compiler.ext}"
    new_contract_1.write_text("foobar2", encoding="utf8")

    expected_0 = "this is expected message 0"
    expected_1 = "this is expected message 1"

    class MyCustomCompilerError0(CompilerError):
        def __init__(self):
            super().__init__(expected_0)

    class MyCustomCompilerError1(CompilerError):
        def __init__(self):
            super().__init__(expected_1)

    mock_compiler.compile.side_effect = MyCustomCompilerError0
    second_mock_compiler.compile.side_effect = MyCustomCompilerError1
    _ = compilers.registered_compilers  # Ensures cached property is set.
    # Hack in our mock compilers.
    compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler
    compilers.__dict__["registered_compilers"][second_mock_compiler.ext] = second_mock_compiler

    try:
        match = rf"{expected_0}\n\n{expected_1}"
        with pytest.raises(CompilerError, match=match):
            list(compilers.compile([new_contract_0, new_contract_1], project=project_with_contract))

    finally:
        for ext in (mock_compiler.ext, second_mock_compiler.ext):
            if ext in compilers.__dict__.get("registered_compilers", {}):
                del compilers.__dict__["registered_compilers"][ext]


def test_compile_source(compilers):
    code = '[{"name":"foo","type":"fallback", "stateMutability":"nonpayable"}]'
    actual = compilers.compile_source("ethpm", code)
    assert isinstance(actual, ContractContainer)


def test_compile_in_project_where_source_id_matches_local_project(project, compilers):
    """
    Tests against a bug where if you had two projects with the same source IDs but
    different content, it always compiled the local project's source.
    """
    new_abi = {
        "inputs": [],
        "name": "retrieve",
        "outputs": [{"internalType": "uint256", "name": "", "type": "uint256"}],
        "stateMutability": "view",
        "type": "function",
    }
    content = json.dumps([new_abi])
    with project.isolate_in_tempdir() as temp_project:
        assert compilers.local_project.path != temp_project.path, (
            "Cannot be same as local for this test"
        )
        contract = temp_project.load_contracts()["Interface"]
        path = temp_project.sources.lookup(contract.source_id)
        path.unlink(missing_ok=True)
        path.write_text(content, encoding="utf8")

        # NOTE: Another condition for this bug is that the given path
        #   must be in source-ID form, meaning it relative to the project
        #   (but not _necessarily_ a relative path, e.g. no `./` prefix).
        argument = Path(contract.source_id)

        # Compile the file with the same name but different content.
        result = [
            x
            for x in compilers.compile([argument], project=temp_project)
            if x.name == contract.name
        ][0]

        # It should reflect the new content and not the one with the same
        # source ID from the local project.
        assert "retrieve" in result.methods


def test_enrich_error_custom_error(chain, compilers):
    abi = [ErrorABI(type="error", name="InsufficientETH", inputs=[])]
    contract_type = ContractType(abi=abi)
    addr = cast(AddressType, "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD")
    err = ContractLogicError("0x6a12f104", contract_address=addr)

    # Hack in contract-type.
    chain.contracts.contract_types[addr] = contract_type

    # Enriching the error should produce a custom error from the ABI.
    actual = compilers.enrich_error(err)

    assert isinstance(actual, CustomError)
    assert actual.__class__.__name__ == "InsufficientETH"


def test_enrich_error_custom_error_with_inputs(compilers, setup_custom_error):
    deadline = 5
    address = "0x3fC91A3afd70395Cd496C647d5a6CC9D4B2b7FAD"
    err = ContractLogicError(
        f"0xd81b2f2e000000000000000000000000000000000000000000000000000000000000000{deadline}",
        contract_address=cast(AddressType, address),
    )
    setup_custom_error(address)

    # Enriching the error should produce a custom error from the ABI.
    actual = compilers.enrich_error(err)

    assert isinstance(actual, CustomError)
    assert actual.__class__.__name__ == "AllowanceExpired"
    assert actual.inputs["deadline"] == deadline
    assert repr(actual) == f"AllowanceExpired(deadline={deadline})"


def test_config_exclude_regex_serialize():
    """
    Show we can to-and-fro with exclude regexes.
    """
    raw_value = 'r"FooBar"'
    cfg = Config(exclude=[raw_value])
    excl = [x for x in cfg.exclude if isinstance(x, Pattern)]
    assert len(excl) == 1
    assert excl[0].pattern == "FooBar"
    # NOTE: Use json mode to ensure we can go from most minimum value back.
    model_dump = cfg.model_dump(mode="json", by_alias=True)
    assert raw_value in model_dump.get("exclude", [])
    new_cfg = Config.model_validate(cfg.model_dump(mode="json", by_alias=True))
    excl = [x for x in new_cfg.exclude if isinstance(x, Pattern)]
    assert len(excl) == 1
    assert excl[0].pattern == "FooBar"



from tests.conftest import geth_process_test


@geth_process_test
def test_append_gas(gas_tracker, geth_account, geth_contract):
    tx = geth_contract.setNumber(924, sender=geth_account)
    trace = tx.trace
    gas_tracker.append_gas(trace, geth_contract.address)
    report = gas_tracker.session_gas_report
    contract_name = geth_contract.contract_type.name
    assert contract_name in report
    assert "setNumber" in report[contract_name]
    assert tx.gas_used in report[contract_name]["setNumber"]


@geth_process_test
def test_append_gas_deploy(gas_tracker, geth_contract):
    tx = geth_contract.creation_metadata.receipt
    trace = tx.trace
    gas_tracker.append_gas(trace, geth_contract.address)
    report = gas_tracker.session_gas_report
    expected = {geth_contract.contract_type.name: {"__new__": [tx.gas_used]}}
    assert report == expected



from typing import TYPE_CHECKING, Optional

import pytest

from ape.api.explorers import ExplorerAPI

if TYPE_CHECKING:
    from ethpm_types import ContractType

    from ape.types.address import AddressType


class MyExplorer(ExplorerAPI):
    def get_transaction_url(self, transaction_hash: str) -> str:
        return ""

    def get_address_url(self, address: "AddressType") -> str:
        return ""

    def get_contract_type(self, address: "AddressType") -> Optional["ContractType"]:
        return None

    def publish_contract(self, address: "AddressType"):
        return


@pytest.fixture
def explorer(networks):
    return MyExplorer(name="mine", network=networks.ethereum.local)


def test_supports_chain(explorer):
    assert not explorer.supports_chain(1)



from pathlib import Path

import pytest
from ethpm_types import MethodABI
from ethpm_types.source import ContractSource, Source

import ape
from ape.pytest.config import ConfigWrapper
from ape.pytest.coverage import CoverageData, CoverageTracker
from ape.types.coverage import (
    ContractCoverage,
    ContractSourceCoverage,
    CoverageProject,
    CoverageReport,
    CoverageStatement,
    FunctionCoverage,
)
from ape.types.trace import SourceTraceback

STMT_0_HIT = 12
STMT_1_HIT = 9
STMT_2_HIT = 0


@pytest.fixture
def statements():
    return create_statements(20, 21, 21)


def create_statements(*pcs) -> list[CoverageStatement]:
    return [
        CoverageStatement(pcs={pcs[0]}, hit_count=STMT_0_HIT),
        CoverageStatement(pcs={pcs[1]}, hit_count=STMT_1_HIT),
        CoverageStatement(pcs={pcs[2]}, hit_count=STMT_2_HIT),
    ]


@pytest.fixture
def foo_function(statements):
    return FunctionCoverage(name="foo", full_name="foo()", statements=statements, hit_count=1)


@pytest.fixture
def bar_function():
    return FunctionCoverage(
        name="bar", full_name="bar()", statements=create_statements(30, 31, 32), hit_count=0
    )


@pytest.fixture
def function(foo_function):
    return foo_function  # alias for when there's only 1 function in test.


@pytest.fixture
def contract(foo_function, bar_function):
    return ContractCoverage(name="Contract", functions=[foo_function, bar_function])


@pytest.fixture
def second_contract(foo_function, bar_function):
    # It's ok to have the same functions.
    return ContractCoverage(name="Contract_Second", functions=[foo_function, bar_function])


@pytest.fixture
def source_contract(contract):
    return ContractSourceCoverage(source_id="Contract.vy", contracts=[contract])


@pytest.fixture
def second_source_contract(second_contract):
    return ContractSourceCoverage(source_id="Contract_Second.vy", contracts=[second_contract])


@pytest.fixture
def coverage_project(source_contract, second_source_contract):
    return CoverageProject(name="__local__", sources=[source_contract, second_source_contract])


@pytest.fixture
def coverage_report(coverage_project):
    return CoverageReport(source_folders=[Path.cwd()], projects=[coverage_project], timestamp=0)


class TestFunctionCoverage:
    def test_hit_count(self, function):
        assert function.hit_count == 1

    def test_lines_coverage(self, function):
        assert function.lines_covered == 2

    def test_miss_count(self, function):
        assert function.miss_count == 1  # stmt 3

    def test_line_rate(self, function):
        assert function.line_rate == 2 / 3

    def test_line_rate_when_no_statements(self):
        """
        An edge case: when a function has no statements, its line rate
        it either 0% if it was not called or 100% if it called.
        """
        function = FunctionCoverage(name="bar", full_name="bar()")
        assert function.hit_count == 0
        function.hit_count += 1
        assert function.line_rate == 1


class TestContractCoverage:
    def test_function_rate(self, contract):
        assert contract.function_rate == 0.5

    def test_lines_coverage(self, contract):
        assert contract.lines_covered == 4

    def test_miss_count(self, contract):
        assert contract.miss_count == 2

    def test_line_rate(self, contract):
        assert contract.line_rate == 2 / 3


class TestSourceCoverage:
    def test_function_rate(self, source_contract):
        assert source_contract.function_rate == 0.5

    def test_lines_coverage(self, source_contract):
        assert source_contract.lines_covered == 4

    def test_miss_count(self, source_contract):
        assert source_contract.miss_count == 2

    def test_line_rate(self, source_contract):
        assert source_contract.line_rate == 2 / 3


class TestCoverageProject:
    def test_function_rate(self, coverage_project):
        assert coverage_project.function_rate == 0.5

    def test_lines_coverage(self, coverage_project):
        # Doubles because has 2 sources in it now (with same amounts of things)
        assert coverage_project.lines_covered == 8

    def test_miss_count(self, coverage_project):
        assert coverage_project.miss_count == 4

    def test_line_rate(self, coverage_project):
        assert coverage_project.line_rate == 2 / 3


class TestCoverageReport:
    def test_function_rate(self, coverage_report):
        assert coverage_report.function_rate == 0.5

    def test_lines_coverage(self, coverage_report):
        assert coverage_report.lines_covered == 8

    def test_miss_count(self, coverage_report):
        assert coverage_report.miss_count == 4

    def test_line_rate(self, coverage_report):
        assert coverage_report.line_rate == 2 / 3


class TestCoverageData:
    @pytest.fixture(scope="class")
    def src(self):
        return Source.model_validate("test")

    @pytest.fixture(scope="class")
    def contract_source(self, vyper_contract_type, src):
        return ContractSource(contract_type=vyper_contract_type, source=src)

    @pytest.fixture
    def coverage_data(self, project, contract_source):
        return CoverageData(project, (contract_source,))

    def test_report(self, coverage_data):
        actual = coverage_data.report
        assert isinstance(actual, CoverageReport)


class TestCoverageTracker:
    @pytest.fixture
    def pytest_config(self, mocker):
        return mocker.MagicMock()

    @pytest.fixture
    def config_wrapper(self, pytest_config):
        return ConfigWrapper(pytest_config)

    @pytest.fixture
    def tracker(self, pytest_config, project):
        return CoverageTracker(pytest_config, project=project)

    def test_data(self, tracker):
        assert tracker.data is not None
        actual = tracker.data.project
        expected = tracker.local_project
        assert actual == expected

    def test_cover(self, mocker, pytest_config, compilers, mock_compiler):
        """
        Ensure coverage of a call works.
        """
        filestem = "atest"
        filename = f"{filestem}.__mock__"
        fn_name = "_a_method"

        # Set up the mock compiler.
        mock_compiler.abi = [MethodABI(name=fn_name)]
        mock_compiler.ast = {
            "src": "0:112:0",
            "name": filename,
            "end_lineno": 7,
            "lineno": 1,
            "ast_type": "Module",
        }
        mock_compiler.pcmap = {"0": {"location": (1, 7, 1, 7)}}
        mock_contract = mocker.MagicMock()
        mock_contract.name = filename
        mock_statement = mocker.MagicMock()
        mock_statement.pcs = {20}
        mock_statement.hit_count = 0
        mock_function = mocker.MagicMock()
        mock_function.name = fn_name
        mock_function.statements = [mock_statement]
        mock_contract.functions = [mock_function]
        mock_contract.statements = [mock_statement]

        def init_profile(source_cov, src):
            source_cov.contracts = [mock_contract]

        mock_compiler.init_coverage_profile.side_effect = init_profile

        stmt = {"type": "dev: Cannot send ether to non-payable function", "pcs": [20]}
        fn_name = "_a_method"
        tb_data = {
            "statements": [stmt],
            "closure": {"name": fn_name, "full_name": f"{fn_name}()"},
            "depth": 0,
        }

        with ape.Project.create_temporary_project() as tmp:
            # Create a source file.
            file = tmp.path / "contracts" / filename
            file.parent.mkdir(exist_ok=True, parents=True)
            file.write_text("testing", encoding="utf8")

            # Ensure the TB refers to this source.
            tb_data["source_path"] = f"{tmp.path}/contracts/{filename}"
            call_tb = SourceTraceback.model_validate([tb_data])

            try:
                # Hack in our mock compiler.
                _ = compilers.registered_compilers  # Ensure cache exists.
                compilers.__dict__["registered_compilers"][mock_compiler.ext] = mock_compiler

                # Ensure our coverage tracker is using our new tmp project w/ the new src
                # as well is set _after_ our new compiler plugin is added.
                tracker = CoverageTracker(pytest_config, project=tmp)

                tracker.cover(call_tb, contract=filestem, function=f"{fn_name}()")
                assert mock_statement.hit_count > 0

            finally:
                if (
                    "registered_compilers" in compilers.__dict__
                    and mock_compiler.ext in compilers.__dict__["registered_compilers"]
                ):
                    del compilers.__dict__["registered_compilers"][mock_compiler.ext]



from collections import namedtuple

from eth_utils import to_hex

DEFAULT_NUMBER_OF_TEST_ACCOUNTS = 10
DEFAULT_TEST_MNEMONIC = "test test test test test test test test test test test junk"
DEFAULT_TEST_HD_PATH = "m/44'/60'/0'/0"
DEFAULT_TEST_CHAIN_ID = 1337
DEFAULT_TEST_ACCOUNT_BALANCE = int(10e21)  # 10,000 Ether (in Wei)
GeneratedDevAccount = namedtuple("GeneratedDevAccount", ("address", "private_key"))
"""
An account key-pair generated from the test mnemonic. Set the test mnemonic
in your ``ape-config.yaml`` file under the ``test`` section. Access your test
accounts using the :py:attr:`~ape.managers.accounts.AccountManager.test_accounts` property.

Config example::

    test:
      mnemonic: test test test test test test test test test test test junk
      number_of_accounts: 10

"""


def generate_dev_accounts(
    mnemonic: str = DEFAULT_TEST_MNEMONIC,
    number_of_accounts: int = DEFAULT_NUMBER_OF_TEST_ACCOUNTS,
    hd_path: str = DEFAULT_TEST_HD_PATH,
    start_index: int = 0,
) -> list[GeneratedDevAccount]:
    """
    Create accounts from the given test mnemonic.
    Use these accounts (or the mnemonic) in chain-genesis
    for testing providers.

    Args:
        mnemonic (str): mnemonic phrase or seed words.
        number_of_accounts (int): Number of accounts. Defaults to ``10``.
        hd_path(str): Hard Wallets/HD Keys derivation path format.
          Defaults to ``"m/44'/60'/0'/0"``.
        start_index (int): The index to start from in the path. Defaults
          to 0.

    Returns:
        list[:class:`~ape.utils.GeneratedDevAccount`]: List of development accounts.
    """
    # perf: lazy imports so module loads faster.
    from eth_account.hdaccount.mnemonic import Mnemonic

    seed = Mnemonic.to_seed(mnemonic)
    hd_path_format = (
        hd_path if "{}" in hd_path or "{0}" in hd_path else f"{hd_path.rstrip('/')}/{{}}"
    )
    return [
        _generate_dev_account(hd_path_format, i, seed)
        for i in range(start_index, start_index + number_of_accounts)
    ]


def _generate_dev_account(hd_path, index: int, seed: bytes) -> GeneratedDevAccount:
    # perf: lazy imports so module loads faster.
    from eth_account.account import Account
    from eth_account.hdaccount import HDPath

    return GeneratedDevAccount(
        address=Account.from_key(
            private_key := to_hex(HDPath(hd_path.format(index)).derive(seed))
        ).address,
        private_key=private_key,
    )



def __getattr__(name: str):
    import ape.contracts.base as module

    return getattr(module, name)


__all__ = [
    "ContractContainer",
    "ContractEvent",
    "ContractInstance",
    "ContractLog",
    "ContractNamespace",
]
